/*
 * This file is part of the segmentation and registration toolkit (SphereSegRegTk).
 *
 * SphereSegRegTk is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * SphereSegRegTk is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with SphereSegRegTk.  If not, see <http://www.gnu.org/licenses/>.
 */
#ifndef CFILTERHANDLER_HPP
#define CFILTERHANDLER_HPP

#include <iostream>
#include <fstream>
#include <sys/time.h>
#include "libseg/CData.hpp"

#include "CDicomImage.hpp"
#include "libseg/CFilterSphere.hpp"
#include "libseg/CFilterGradient.hpp"
#include "libseg/CFilterParticles.hpp"
#include "libseg/CFilterGraph.hpp"
#include "libseg/CFilterMatchingSimple.hpp"
#include "libseg/CDirectory.hpp"
#include "libseg/CFilterProjection.hpp"


/// Handler class to fuse all the filter to one class
template <typename T>
class CFilterHandler
{
public:
	/**
	 * type of variables which can be stored as parameters
	 */
	enum SParamType
	{
		FLOAT,
		INT,
		BOOL,
		STRING
	};

	/// filter parameter handling (store/load)
	/**
	 * parameter class to handle storing and loading of parameter values of gui
	 */
	class SParam
	{
	public:
		// parameter name
		const char *name;

		// type of value
		SParamType type;

		// value for float parameter
		// initialize with nan which will be overwritten on loading with a non nan number (even if loading int) for error detection
		float float_value;
		int int_value;
		string string_value;
		bool bool_value;
	};

	CFilterHandler();
	~CFilterHandler();

public:
	std::string lastErrorMessage;	///< error message for debugging and gui purposes

	// RAW DATA
	std::string raw_filename;	///< filename (e.g. to reset data)
	bool raw_valid;			///< raw data is valid
	CData<T> raw_data;		///< image data
	CVector<2,T> raw_limits;	///< minimum and maximum values in raw image

	// SPHERE DATA
	CFilterSphere<T> sphere_filter;
	bool apply_sphere_filter();	///< apply the sphere filter
	bool apply_sphere_preview();	///< generate a preview (threshold using sphere_limits)
	bool sphere_valid;		///< sphere data valid
	CVector<2,T> sphere_limits;	///< limits (min/max values) of sphere data
	CData<T> sphere_data;		///< sphere data

	// GRADIENT DATA
	CFilterGradient<T> gradient_filter;
	bool apply_gradient_filter();	///< apply the gradient filter
	bool gradient_valid;		///< gradient data is valid
	CVector<2,T> gradient_limits;
	CData< CVector<3,T> > gradient_data;	///< gradient data

	// PARTICLE FILTER
	CFilterParticles<T> particles_filter;
	bool particles_valid;		///< particle data valid
	bool apply_particles_filter();	///< apply the particle filter

	// GRAPH FILTER
	CFilterGraph<T> graph_filter;
	bool apply_graph_filter();	///< apply the graph filter
	bool graph_valid;		///< graph created

	// MATCHING FILTER
	CGraphRaw<3,T> reference_graph_raw;
	CGraph<3,T> reference_graph;	///< reference graph
	bool reference_graph_valid;	///< graph to match with graph from previous filter is loaded and ready for matching

	// MATCHING SIMPLE FILTER - match reference_graph to the computed graph!
	CFilterMatchingSimple<T> matching_simple_filter;
	bool apply_matching_simple_filter();	///< apply the matching
	bool matching_simple_valid;		///< matching was performed and matching data is available

	// PROJECTION DATA
	CFilterProjection<T> transformation_filter;
	bool apply_transformation_filter();	///< apply the transformation
	bool transformation_valid;		///< destination image was projected
	CData<T> transformation_data;		///< projected data
	CVector<2,T> transformation_limits;	///< minimum and maximum values in transformation image

	// RAW DATA AFTER TRANSFORMATION
	std::string reference_filename;		///< filename (e.g. to reset data)
	bool reference_valid;
	CData<T> reference_data;		///< image data for second graph
	CVector<2,T> reference_limits;		///< minimum and maximum values in reference image
public:

	/**
	 * reset the data (empty data containers)
	 * set all 'valid' variables to false
	 */
	void reset();


	/**
	 * Save the parameters to the filename of the [reference filename] + [special automagically created extensions]
	 */
	void save(SParam *params = NULL);

private:
	void save_as(std::string& orig_filename, SParam *params = NULL);

public:


	/**
	 * load parameters from [filename] + [special automagically created extensions]
	 */
	bool load_as(std::string& filename, SParam *params = NULL);

	/**
	 * load parameters from [reference filename] + [special automagically created extensions]
	 */
	bool load(SParam *params = NULL);

	/**
	 * load reference graph for matching
	 */
	bool load_graph_raw(std::string& filename);

	/**
	 * scale a dataset to its half size
	 * this can be used if there is not enough memory to handle 512x512x512 CT data
	 * the scaling is done by averaging 2x2x2 pixels
	 */
	void scale_to_half(CData<T> &data);

	/**
	 * scale raw data to half size
	 */
	void scale_raw_to_half();

	/**
	 * scale refernce data to half size
	 */
	void scale_reference_to_half();

	/**
	 * open image given by filename [file] and store it to raw_data.
	 * if the directory contains multiple images, load whole images (e.g. multiple ct slices).
	 *
	 * the dicom headers are not read for the correct alignment. it's assumed that the ct dataslices
	 * are enumerated correctly by the filename
	 *
	 * \param filepath	one filename with path of dataset
	 * \param raw_data	storage for image data
	 * \param raw_limits	min/max values of image data
	 * \param raw_filename	raw_filename to store unique filename for whole dataset
	 */
	bool open_raw(std::string& filepath, CData<T> &raw_data, CVector<2,T> &raw_limits, string &raw_filename)
	{
		// extract path of filepath
		string path;
		size_t pos = filepath.rfind('/');
		if (pos == filepath.npos)	path = "./";	// use current path if no delimiter was found
		else				path = filepath.substr(0, pos);

		string file_name = filepath.substr(pos+1);	// add 1 because of '/'

		// find first file_name letters - the format is [some letters](\.[number]).IMA
		size_t file_name_length = file_name.length();

		if (file_name_length < 4)	// the filelength has to be at least 4
			return false;

		// COPY REVERSE version of filename (without the .IMA extension)
		string file_begin;
		size_t file_begin_length = file_name.length() - 4;
		file_begin.resize(file_begin_length);

		for (size_t i = 0; i < file_begin_length; i++)
			file_begin[file_begin_length - i - 1] = file_name[i];

		// find first occurence of a character not [.] or [number]
		size_t fixed_part_pos = file_begin.find_first_not_of(".1234567890", 0);

		if (fixed_part_pos == string::npos)
		{
			std::cout << "number only dicom file" << std::endl;
			fixed_part_pos = file_begin.length();

			file_begin = "";
			file_begin_length = file_begin.length();
		}
		else
		{
			file_begin = file_begin.substr(fixed_part_pos);
			file_begin_length = file_begin.length();
		}

		// reverse again to gain the filename without the ending numbers and extension
		for (size_t i = 0; i < file_begin_length/2; i++)
		{
			char a = file_begin[i];
			file_begin[i] = file_begin[file_begin_length-i-1];
			file_begin[file_begin_length-i-1] = a;
		}

		std::cout << "File starts with: \"" << file_begin << "\"" << std::endl;

		// load all files within directory
		CDirectory dir;
		list<string> path_files;
		dir.getFiles(path_files, path);

		// filter out files not ending with .IMA
		list<string> dicom_files;

		for (list<string>::iterator i = path_files.begin(); i != path_files.end(); i++)
		{
			string &file = *i;
			if (file.length() < 4)
				continue;
			string file_ending = file.substr(file.length()-4, 4);

			if (file_ending != ".IMA" && file_ending != ".ima" && file_ending != ".dcm")
				continue;

			if (file_begin != file.substr(0, file_begin_length))
				continue;
			dicom_files.push_back(file);
		}

		// depth of the new CT dataset is the number of dicom files
		int depth = dicom_files.size();
		if (depth == 0)
		{
			lastErrorMessage = "No dicom files ending with .ima or .IMA found";
			return false;
		}

		dicom_files.sort();

		std::cout << "loading " << depth << " dicom file(s)..." << std::endl;

		int slice_id = 0;
		int width = -1;
		int height = -1;
		float min_value = 0;
		float max_value = 0;
		for (list<string>::iterator i = dicom_files.begin(); i != dicom_files.end(); i++)
		{
			string file = *i;	// file within directory
			string file_with_path = path + "/" + *i;	// construct filepath

			std::cout << "loading file '" << file << "'" << std::endl;

			// load dicom file
			CDicomImage dcImage(file_with_path);
			if (!dcImage)
			{
				std::ostringstream buf;
				buf << "unable to load file: " << path << "/" << *i << std::endl;
				lastErrorMessage = buf.str();
				return false;
			}
			else
			{
				std::ostringstream buf;
				if (width == -1)
				{
					// ct data property variables are not initialized
					min_value = dcImage.min_value;
					max_value = dcImage.max_value;
					width = dcImage.width;
					height = dcImage.height;

					// allocate space
					raw_data.resize(width, height, depth);
				}
				else
				{
					// update window border values
					if (min_value > dcImage.min_value)
						min_value = dcImage.min_value;
					if (max_value < dcImage.max_value)
						max_value = dcImage.max_value;

					// check that all slices have the same resolution
					if (width != dcImage.width || height != dcImage.height)
					{
						std::ostringstream buf;
						buf << "last dicom image has different resolution (" << width << " " << height << ") != (" << dcImage.width << " " << dcImage.height << ") ... aborting" << std::endl;
						std::cerr << buf.str() << std::endl;
						lastErrorMessage = buf.str();
						return false;
					}
				}

				// load data of slice to raw_data
				raw_data.loadDataSlice(dcImage.data, slice_id);
			}
			slice_id++;
		}

		raw_limits[0] = min_value;
		raw_limits[1] = max_value;

		// use first filename of sorted dicom files as raw_filename to get unique file storage for whole ct data
		raw_filename = path + "/" + dicom_files.front();

		std::cout << "files loaded with raw filename: " << raw_filename << "  min/max: " << raw_limits << std::endl;

		return true;
	}

	/**
	 * open raw data to apply filters on it
	 */
	bool open_raw_data(std::string& file)
	{
		if (!open_raw(file, raw_data, raw_limits, raw_filename))
			return false;

		raw_valid = true;
		return true;
	}

	/**
	 * open reference data to compare it with transformed raw_data
	 */
	bool open_reference_data(std::string& file)
	{
		if (!open_raw(file, reference_data, reference_limits, reference_filename))
			return false;

		reference_valid = true;
		return true;
	}

	bool output_timings;	// output the time taken by each filter

private:
	double last_timestamp;	// timestamp taken with start_take_time()

	/**
	 * take the current timestamp
	 */
	void start_take_time()
	{
		if (!output_timings)	return;

		timeval now;
		gettimeofday(&now, NULL);
		last_timestamp = (double)now.tv_usec/(double)1000000.0;
		last_timestamp += (double)now.tv_sec;
	}

	/**
	 * output the difference between the timestamp taken with start_take_time() and the current time
	 */
	void end_take_time()
	{
		if (!output_timings)	return;

		double new_timestamp;
		timeval now;
		gettimeofday(&now, NULL);
		new_timestamp = (double)now.tv_usec/(double)1000000.0;
		new_timestamp += (double)now.tv_sec;

		std::cout << "Seconds: " << (new_timestamp - last_timestamp) << std::endl;
	}
};



template <typename T>
CFilterHandler<T>::~CFilterHandler()
{
}


template <typename T>
void CFilterHandler<T>::scale_to_half(CData<T> &data)
{
	CData<float> new_data(data.w/2, data.h/2, data.d/2);

	for (int z = 0; z < new_data.d; z++)
		for (int y = 0; y < new_data.h; y++)
			for (int x = 0; x < new_data.w; x++)
			{
				new_data.set(x, y, z,
					(
						data.get(x*2+0, y*2+0, z*2+0)	+	// average with 2x2x2 kernel
						data.get(x*2+1, y*2+0, z*2+0)	+
						data.get(x*2+0, y*2+1, z*2+0)	+
						data.get(x*2+1, y*2+1, z*2+0)	+
						data.get(x*2+0, y*2+0, z*2+1)	+
						data.get(x*2+1, y*2+0, z*2+1)	+
						data.get(x*2+0, y*2+1, z*2+1)	+
						data.get(x*2+1, y*2+1, z*2+1)
					) * ((T)1/(T)8));
			}

	data.resize(new_data.w, new_data.h, new_data.d);

	// copy back
	for (size_t i = 0; i < data.pixel_size; i++)
		data.data[i] = new_data.data[i];
}


template <typename T>
void CFilterHandler<T>::scale_raw_to_half()
{
	if (!raw_valid)
		return;

	scale_to_half(raw_data);
}


template <typename T>
void CFilterHandler<T>::scale_reference_to_half()
{
	if (!reference_valid)
		return;

	scale_to_half(reference_data);
}


/**
 * reset the class to it's initial values
 */
template <typename T>
void CFilterHandler<T>::reset()
{
	output_timings = true;

	raw_valid = false;
	raw_data.reset();

	sphere_valid = false;
	sphere_data.reset();

	gradient_valid = false;
	gradient_data.reset();

	particles_valid = false;
	particles_filter.reset();

	graph_valid = false;
	// TODO
	//graph_filter.reset();

	reference_graph_valid = false;	// graph to match is loaded

	matching_simple_valid = false;	// simple filter was applied

	transformation_valid = false;
	transformation_data.reset();

	reference_valid = false;
	reference_data.reset();
}




template <typename T>
bool CFilterHandler<T>::apply_sphere_filter()
{
	if (!raw_valid)
	{
		lastErrorMessage = "No raw data available";
		return false;
	}

	// apply filter
	start_take_time();
	sphere_filter.filter(raw_data, sphere_data, &last_timestamp);
	end_take_time();

	sphere_limits[0] = sphere_data.getMin();
	sphere_limits[1] = sphere_data.getMax();

	std::cout << sphere_limits[0] << " " << sphere_limits[1] << std::endl;

	sphere_valid = true;

	return true;
}



template <typename T>
bool CFilterHandler<T>::apply_sphere_preview()
{
	if (!raw_valid)
	{
		lastErrorMessage = "No raw data available";
		return false;
	}

	// apply filter
	sphere_filter.filter_preview(raw_data, sphere_data);

	sphere_limits[0] = sphere_data.getMin();
	sphere_limits[1] = sphere_data.getMax();

	sphere_valid = true;

	return true;
}


template <typename T>
bool CFilterHandler<T>::apply_gradient_filter()
{
	if (!sphere_valid)
	{
		lastErrorMessage = "No spherical data available";
		return false;
	}

	// apply filter
	start_take_time();
	gradient_filter.filter(sphere_data, gradient_data);
	end_take_time();

	gradient_limits[0] = gradient_data.getMin();
	gradient_limits[1] = gradient_data.getMax();

	gradient_valid = true;

	return true;
}



template <typename T>
bool CFilterHandler<T>::apply_particles_filter()
{
	if (!gradient_valid)
	{
		lastErrorMessage = "No gradient data available";
		return false;
	}

	if (gradient_data.w != sphere_data.w || gradient_data.h != sphere_data.h)
	{
		lastErrorMessage = "Size of gradient and sphere data does not match";
		return false;
	}

	// apply filter
	start_take_time();
	particles_filter.filter(gradient_data, sphere_data);
	end_take_time();

	particles_valid = true;

	return true;
}



template <typename T>
bool CFilterHandler<T>::apply_graph_filter()
{
	if (!particles_valid)
	{
		lastErrorMessage = "No particle data available";
		return false;
	}

	// apply filter
	start_take_time();
	graph_filter.filter(	particles_filter.particleList,
				particles_filter.particle_domain,
				particles_filter,
				gradient_data
			);
	end_take_time();

	graph_valid = true;

	return true;
}



template <typename T>
bool CFilterHandler<T>::apply_matching_simple_filter()
{

	if (!graph_valid)
	{
		lastErrorMessage = "No graph data (from previous filters) available";
		return false;
	}

	if (!reference_graph_valid)
	{
		lastErrorMessage = "No matching graph data available";
		return false;
	}

	/* apply filter
	 * With the next filter, we want to reproject the original image to the space of the other graph (reference_graph).
	 * To avoid inverting the matrix, we create a matrix, projecting the reference_graph to the just computed graph.
	 * This makes it possible to iterate through the domain of reference_graph using the transformation matrix to compuate
	 * the according point in the original space
	 */
	start_take_time();
	matching_simple_filter.filter(reference_graph, graph_filter.graph);
	end_take_time();

	matching_simple_valid = true;

	return true;
}


template <typename T>
bool CFilterHandler<T>::apply_transformation_filter()
{

	if (!raw_valid)
	{
		lastErrorMessage = "No raw data available";
		return false;
	}

	if (!reference_valid)
	{
		lastErrorMessage = "No raw data of second graph available - please load it";
		return false;
	}

	// apply filter
	CVector<3,int> size(reference_data.w, reference_data.h, reference_data.d);
	start_take_time();
	transformation_filter.filter(raw_data, transformation_data, size);
	end_take_time();

	transformation_limits = raw_limits;
	transformation_valid = true;

	return true;
}


template <typename T>
void CFilterHandler<T>::save(SParam *params)
{
	save_as(raw_filename, params);
}


/**
 * store the computed data
 * \param orig_filename	original filename to construct separate filenames for each filter result
 * \param param		parameters to store
 * 			set to NULL to disable storing parameters
 * 			the list is terminated by setting name to NULL
 */
template <typename T>
void CFilterHandler<T>::save_as(std::string& orig_filename, SParam *params)
{
	string filename;
	ofstream file;
	CMemoryContainer container;

	if (sphere_valid)
	{
		filename = orig_filename + "_sphere.sav";
		file.open(filename.c_str(), std::ios::binary | std::ios::trunc);
			CFilterSphere<T>::save(container, sphere_data);

		file.write((char*)container.data(), container.size());
		file.close();
		container.clear();
	}

	if (gradient_valid)
	{
		filename = orig_filename + "_gradient.sav";
		file.open(filename.c_str(), std::ios::binary | std::ios::trunc);
		CFilterGradient<T>::save(container, gradient_data);
		file.write((char*)container.data(), container.size());
		file.close();
		container.clear();
	}

	if (particles_valid)
	{
		filename = orig_filename + "_particles.sav";
		file.open(filename.c_str(), std::ios::binary | std::ios::trunc);
		CFilterParticles<T>::save(particles_filter.particleList, container, particles_filter.particle_domain);
		file.write((char*)container.data(), container.size());
		file.close();
		std::cout << filename << " saved (" << particles_filter.particleList.size() << ")" << std::endl;
		container.clear();
	}

	if (graph_valid)
	{
		filename = orig_filename + "_graph_raw.sav";
		file.open(filename.c_str(), std::ios::binary | std::ios::trunc);
		CFilterGraph<T>::save(graph_filter.strips, container);
		file.write((char*)container.data(), container.size());
		file.close();
		std::cout << "   saved" << std::endl;
		container.clear();
	}

	if (params != NULL)
	{
		filename = orig_filename + "_parameters.sav";
		std::ofstream pfile;
		pfile.open(filename.c_str(), std::ios::trunc);

		while(params->name != NULL)
		{
			pfile << params->name << "\t";
			switch(params->type)
			{
				case FLOAT:	pfile << params->float_value;	break;
				case INT:	pfile << params->int_value;	break;
				case BOOL:	pfile << params->bool_value;	break;
				case STRING:	pfile << params->string_value;	break;
			}
			pfile << std::endl;

			params++;
		}

		pfile.close();
	}
}


template <typename T>
bool CFilterHandler<T>::load(SParam *params)
{
	return load_as(raw_filename, params);
}


template <typename T>
bool CFilterHandler<T>::load_as(std::string& orig_filename, SParam *params)
{
	string filename;
	ifstream file;
	CMemoryContainer container;

	filename = orig_filename + "_sphere.sav";
	file.open(filename.c_str(), ios::binary);
	if (!file.is_open())
	{
		std::cout << "[WARNING] failed to open file" << std::endl;
	}
	else
	{
		file.seekg (0, ios::end);
		unsigned int length = file.tellg();
		container.setSize(length);
		file.seekg (0, ios::beg);
		file.read((char*)container.data(), length);
		if (CFilterSphere<T>::load(container, sphere_data))
		{
			sphere_valid = true;
			std::cout << filename << " loaded" << std::endl;
		}
		else
		{
			std::cout << filename << " FAILED" << std::endl;
			sphere_valid = false;
			sphere_data.reset();
		}
		file.close();

		sphere_limits[0] = sphere_data.getMin();
		sphere_limits[1] = sphere_data.getMax();
	}

	filename = orig_filename + "_gradient.sav";
	file.open(filename.c_str(), ios::binary);
	if (!file.is_open())
	{
		std::cout << "[WARNING] failed to open file" << std::endl;
	}
	else
	{
		file.seekg (0, ios::end);
		unsigned int length = file.tellg();
		container.setSize(length);
		file.seekg (0, ios::beg);
		file.read((char*)container.data(), length);
		if (CFilterGradient<T>::load(container, gradient_data))
		{
			gradient_valid = true;
			std::cout << filename << " loaded" << std::endl;
		}
		else
		{
			std::cout << filename << " FAILED" << std::endl;
			gradient_valid = false;
			gradient_data.reset();
		}
		file.close();

		gradient_limits[0] = gradient_data.getMin();
		gradient_limits[1] = gradient_data.getMax();
	}


	filename = orig_filename + "_particles.sav";
	file.open(filename.c_str(), ios::binary);
	if (!file)
	{
		std::cout << "[WARNING] failed to open file" << std::endl;
	}
	else
	{
		file.seekg (0, ios::end);
		unsigned int length = file.tellg();
		container.setSize(length);
		file.seekg (0, ios::beg);
		file.read((char*)container.data(), length);
		if (CFilterParticles<T>::load(particles_filter.particleList, container, particles_filter.particle_domain))
		{
			particles_valid = true;
			std::cout << filename << " loaded" << std::endl;
		}
		else
		{
			particles_valid = false;
			particles_filter.reset();
			std::cout << filename << " FAILED" << std::endl;
		}
		file.close();

		std::cout << filename << " loaded (" << particles_filter.particleList.size() << ")" << std::endl;
	}

	filename = orig_filename + "_graph_raw.sav";
	file.open(filename.c_str(), ios::binary);
	if (!file.is_open())
	{
		std::cout << "[WARNING] failed to open file" << std::endl;
	}
	else
	{
		file.seekg (0, ios::end);
		unsigned int length = file.tellg();
		container.setSize(length);
		file.seekg (0, ios::beg);
		file.read((char*)container.data(), length);
		if (CFilterGraph<T>::load(graph_filter.strips, container))
		{
			graph_valid = true;
			std::cout << filename << " loaded" << std::endl;
		}
		else
		{
			std::cout << filename << " FAILED" << std::endl;
			graph_valid = false;
			//graph_filter.reset();
		}
		file.close();
		std::cout << "   loaded" << std::endl;

		std::cout << "converting strips to graph..." << std::flush;
		graph_filter.strips.toGraph(graph_filter.graph);
		std::cout << std::endl;
	}

	if (params != NULL)
	{
		filename = orig_filename + "_parameters.sav";
		std::ifstream pfile;
		pfile.open(filename.c_str());

		string line;

		while(!pfile.eof())
		{
			// load line
			getline(pfile, line);

			// search for delimiter \t
			size_t found;
			found = line.find("\t");
			if (found==string::npos)
				break;

			// load name + value
			string name = line.substr(0, found);
			string value = line.substr(found+1);

			// search for appropriate value storage
			SParam *tparams = params;
			while(tparams->name != NULL)
			{
				if (name == string(tparams->name))
				{
					tparams->float_value = 0;	// set to some value to clear the error producing nan in float_values

					switch(tparams->type)
					{
						case FLOAT:	tparams->float_value = atof(value.c_str());	break;
						case INT:	tparams->int_value = atoi(value.c_str());	break;
						case BOOL:	tparams->bool_value = atoi(value.c_str());	break;
						case STRING:	tparams->string_value = value;			break;
					}
					break;
				}
				tparams++;
			}

			// if parameter name was not found, we abort
			if (tparams->name == NULL)	// parameter was not found
				break;
		}

		pfile.close();
	}

	return true;
}


template <typename T>
bool CFilterHandler<T>::load_graph_raw(std::string& filename)
{
	ifstream file;

	CMemoryContainer container;
	file.open(filename.c_str(), ios::binary);

	if (!file.is_open())
	{
		std::cout << "[WARNING] failed to open file" << std::endl;
	}
	else
	{
		file.seekg (0, ios::end);
		unsigned int length = file.tellg();
		container.setSize(length);
		file.seekg (0, ios::beg);
		file.read((char*)container.data(), length);
		if (CFilterGraph<T>::load(reference_graph_raw, container))
		{
			reference_graph_valid = true;
			reference_graph_raw.toGraph(reference_graph);
		}
		else
		{
			reference_graph_valid = false;
			std::cout << filename << " FAILED" << std::endl;
		}
		file.close();
		std::cout << "   loaded" << std::endl;
	}

	return reference_graph_valid;
}

template <typename T>
CFilterHandler<T>::CFilterHandler()
{
	reset();
}


#endif
