/*
 * This file is part of the segmentation and registration toolkit (SphereSegRegTk).
 *
 * SphereSegRegTk is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * SphereSegRegTk is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with SphereSegRegTk.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef __CDATA3_HH
#define __CDATA3_HH

#include "CVector.hh"
#include <limits>


/// 3D Data storage container
/**
 * 3D data storage container for interclass communication.
 */
template <typename T>
class CData
{
public:
	int w;	///< width
	int h;	///< height
	int d;	///< depth
	size_t pixel_size;	///< voxels count stored in the dataset (width*height*depth)
	T *data;		///< pointer to data array

	/**
	 * check if a given position is within the allowed border of the storage domain
	 * \param pos	position in the domain
	 */
	inline bool inCheck(CVector<3,int> &pos)
	{
		return(	pos[0] >= 0 &&
			pos[1] >= 0 &&
			pos[2] >= 0 &&
			pos[0] < w &&
			pos[1] < h &&
			pos[2] < d
		);
	}

	/**
	 * check if a given position is within the allowed border of the storage domain
	 * \param x,y,z	position in the domain
	 */
	inline bool inCheck(int x, int y, int z)
	{
		return(	x >= 0 &&
			y >= 0 &&
			z >= 0 &&
			x < w &&
			y < h &&
			z < d
		);
	}

	/**
	 *  update all values of dataset with the given value
	 */
	inline void setValue(T value = (T)0)
	{
		if (data != NULL)
			for (size_t i = 0; i < pixel_size; i++)
				data[i] = value;
	}

	/**
	 * update value at pos with value
	 */
	void inline set(CVector<3,int> pos, T value)
	{
		T *p = data + pos[0] + w*pos[1] + w*h*pos[2];
		*p = value;
	}

	/**
	 * update value at (x,y,z) with value
	 */
	void inline set(int x, int y, int z, T value)
	{
		T *p = data + x + w*y + w*h*z;
		*p = value;
	}

	/**
	 * get value at the position vector
	 */
	inline T &get(CVector<3,int> &pos)
	{
		return *(data + pos[0] + w*pos[1] + w*h*pos[2]);
	}

	/**
	 * get value at position (x,y,z)
	 */
	inline T &get(int x, int y, int z)
	{
		return *(data + x + w*y + w*h*z);
	}

	/**
	 * Get the maximum value of the stored dataarray
	 */
	float getMax();

	/**
	 * Get the minimum value of the stored dataarray
	 */
	float getMin();

	/**
	 * Create new data storage container with given sizes
	 */
	CData(int width, int height, int depth)
	{
		w = width;
		h = height;
		d = depth;
		pixel_size = (size_t)(w*h*d);
		data = new T[pixel_size];
	}

	/**
	 * Resize data storage, the data is not preserved
	 */
	inline void resize(CVector<3,int> &size)
	{
		if (data != NULL)
			delete data;

		w = size[0];
		h = size[1];
		d = size[2];

		pixel_size = (size_t)(w*h*d);
		data = new T[pixel_size];
	}

	/**
	 * Resize data storage, the data is not preserved
	 */
	inline void resize(int p_w, int p_h, int p_d)
	{
		if (data != NULL)
			delete data;

		w = p_w;
		h = p_h;
		d = p_d;
		data = new T[w*h*d];
		pixel_size = (size_t)(w*h*d);
	}

	/**
	 * reset data container and free all data
	 */
	void reset()
	{
		if (data != NULL)
		{
			delete data;
			data = 0;
		}
		w = 0;
		h = 0;
		d = 0;
		pixel_size = 0;
	}

	/**
	 * constructor for empty dataset
	 */
	CData()
	{
		data = NULL;
		reset();
	}


	~CData()
	{
		if (data != NULL)
			delete data;
	}

	/**
	 * Load the red channel of a 24 bpp image to the data storage.
	 */
	void inline loadData24bpp(const unsigned char *src)
	{
		T *dst = data;
		for (size_t i = 0; i < pixel_size; i++)
		{
			*dst = *src;
			src+=3;
			dst++;
		}
	}

	/**
	 * Load the red channel of a 32 bpp image to the data storage.
	 */
	void inline loadData32bpp(const unsigned char *src)
	{
		T *dst = data;
		for (size_t i = 0; i < pixel_size; i++)
		{
			*dst = *src;
			src+=4;
			dst++;
		}
	}


	/**
	 * Multiply each value in the data array with the value "scale"
	 */
	void inline scale(float scale)
	{
		T *dst = data;
		for (size_t i = 0; i < pixel_size; i++)
		{
			*dst *= scale;
			dst++;
		}
	}

	/**
	 * Load the data from "src" to the data array (memcpy)
	 */
	void inline loadData(const T *src)
	{
		memcpy(data, src, sizeof(T)*pixel_size);
	}

	/**
	 * Load the data of one slice given by the pointer "src" to one slice of the data array (memcpy)
	 * \param src	source slice data
	 * \param slice	slice_id to store source data
	 */
	void inline loadDataSlice(const T *src, int slice)
	{
		memcpy(data+slice*w*h, src, sizeof(T)*w*h);
	}


	void inline storeData24bpp(unsigned char *dst, size_t stride);
	void inline storeData24bppSlice(unsigned char *dst, int slice, size_t stride);
	void inline storeData32bpp(unsigned char *dst, size_t stride);

	void inline storeData24bppAndScale(unsigned char *dst, float scale, size_t stride);
	void inline storeData32bppAndScale(unsigned char *dst, float scale, size_t stride);

	void inline storeData24bppAndScale(unsigned char *dst, float min, float max, size_t stride);
	void inline storeData24bppAndScaleSlice(unsigned char *dst, float min, float max, int slice, size_t stride);
	void inline storeData32bppAndScale(unsigned char *dst, float min, float max, size_t stride);

	void inline storeData24bppAddOffset(unsigned char *dst, float offset, size_t stride);
	void inline storeData32bppAddOffset(unsigned char *dst, float offset, size_t stride);
};


template<>
inline float CData<float>::getMax()
{
	float max = -std::numeric_limits<float>::max();
	for (size_t i = 0; i < pixel_size; i++)
		if (data[i] > max)	max = data[i];
	return max;
}

template<>
inline float CData< CVector<3,float> >::getMax()
{
	float max = -std::numeric_limits<float>::max();
	for (size_t i = 0; i < pixel_size; i++)
	{
		if (data[i][0] > max)	max = data[i][0];
		if (data[i][1] > max)	max = data[i][1];
		if (data[i][2] > max)	max = data[i][2];
	}
	return max;
}

template<>
inline float CData<float>::getMin()
{
	float min = std::numeric_limits<float>::max();
	for (size_t i = 0; i < pixel_size; i++)
		if (data[i] < min)	min = data[i];
	return min;
}

template<>
inline float CData< CVector<3,float> >::getMin()
{
	float min = std::numeric_limits<float>::max();
	for (size_t i = 0; i < pixel_size; i++)
	{
		if (data[i][0] < min)	min = data[i][0];
		if (data[i][1] < min)	min = data[i][1];
		if (data[i][2] < min)	min = data[i][2];
	}
	return min;
}



/**
 * Store the dataset to the destination 24bpp image data set. Each color channel
 * is set to the same color given by the value in the dataset
 * \param dst		pointer to destination data storage
 * \param stride	stride if the data should be stored to a pixbuf
 */
template<>
void inline CData<float>::storeData24bpp(unsigned char *dst, size_t stride)
{
	if (stride != 0)
		stride -= w*3;	// just compute the last offset

	float *src = data;
	int x = 0;
	for (size_t i = 0; i < pixel_size; i++)
	{
		unsigned char value = (unsigned char)*src;
		*(dst+0) = value;
		*(dst+1) = value;
		*(dst+2) = value;

		src++;
		dst+=3;
		x++;
		if (x == w)
		{
			dst += stride;
			x = 0;
		}
	}
}


/**
 * Store a slice of the dataset to the destination 24bpp image data set. Each color channel
 * is set to the same color given by the value in the dataset
 * \param dst		pointer to destination data storage
 * \param slice		id of slide to be stored to 'dst'
 * \param stride	stride if the data should be stored to a pixbuf
 */
template<>
void inline CData<float>::storeData24bppSlice(unsigned char *dst, int slice, size_t stride)
{
	if (stride != 0)
		stride -= w*3;	// just compute the last offset

	float *src = data + w*h*slice;
	int x = 0;
	for (int i = 0; i < w*h; i++)
	{
		unsigned char value = (unsigned char)*src;
		*(dst+0) = value;
		*(dst+1) = value;
		*(dst+2) = value;

		src++;
		dst+=3;
		x++;
		if (x == w)
		{
			dst += stride;
			x = 0;
		}
	}
}


/**
 * Store a dataset to the destination 32bpp image data set. Each color and alpha channel
 * is set to the same color given by the value in the dataset
 * \param dst		pointer to destination data storage
 * \param stride	stride if the data should be stored to a pixbuf
 */
template<> void inline CData<float>::storeData32bpp(unsigned char *dst, size_t stride)
{
	if (stride != 0)
		stride -= w*4;	// just compute the last offset

	float *src = data;
	int x = 0;
	for (size_t i = 0; i < pixel_size; i++)
	{
		unsigned char value = (unsigned char)*src;
		*(dst+0) = value;
		*(dst+1) = value;
		*(dst+2) = value;
		*(dst+3) = value;

		src++;
		dst+=4;
		x++;
		if (x == w)
		{
			dst += stride;
			x = 0;
		}
	}
}


/**
 * Store a dataset to the destination 24bpp image data set. Each color channel
 * is set to the same color given by the value in the dataset
 *
 * The parameter offset is added to each value before it's stored
 * \param dst		pointer to destination data storage
 * \param offset	offset to add to the value before storing to destination data
 * \param stride	stride if the data should be stored to a pixbuf
 */
template<>
void inline CData<CVector<3,float> >::storeData24bppAddOffset(unsigned char *dst, float offset, size_t stride)
{
	if (stride != 0)
		stride -= w*3;	// just compute the last offset

	CVector<3,float> *src = data;

	int x = 0;
	for (size_t i = 0; i < pixel_size; i++)
	{
		*(dst++) = (unsigned char)((*src)[0]+offset);
		*(dst++) = (unsigned char)((*src)[1]+offset);
		*(dst++) = (unsigned char)((*src)[2]+offset);
		src++;
		x++;
		if (x == w)
		{
			dst += stride;
			x = 0;
		}
	}
}


/**
 * Store a dataset to the destination 32bpp image data set. Each color and alpha channel
 * is set to the same color given by the value in the dataset
 *
 * The parameter offset is added to each value before it's stored
 * \param dst		pointer to destination data storage
 * \param offset	offset to add to the value before storing to destination data
 * \param stride	stride if the data should be stored to a pixbuf
 */
template<>
void inline CData<CVector<3,float> >::storeData32bppAddOffset(unsigned char *dst, float offset, size_t stride)
{
	if (stride != 0)
		stride -= w*4;	// just compute the last offset

	CVector<3,float> *src = data;

	int x = 0;
	for (size_t i = 0; i < pixel_size; i++)
	{
		*(dst++) = (unsigned char)((*src)[0]+offset);
		*(dst++) = (unsigned char)((*src)[1]+offset);
		*(dst++) = (unsigned char)((*src)[2]+offset);
		*(dst++) = 0;
		src++;
		x++;
		if (x == w)
		{
			dst += stride;
			x = 0;
		}
	}
}


/**
 * Store a dataset to the destination 24bpp image data set. Each color channel
 * is set to the same color given by the value in the dataset
 *
 * The parameter scale is multiplied to each value before it's stored
 * \param dst		pointer to destination data storage
 * \param scale		scale parameter for the value before storing to destination data
 * \param stride	stride if the data should be stored to a pixbuf
 */
template<>
void inline CData<float>::storeData24bppAndScale(unsigned char *dst, float scale, size_t stride)
{
	if (stride != 0)
		stride -= w*3;	// just compute the last offset

	float *src = data;
	int x = 0;
	for (size_t i = 0; i < pixel_size; i++)
	{
		unsigned char value = (unsigned char)((*src)*scale);
		*(dst+0) = value;
		*(dst+1) = value;
		*(dst+2) = value;

		src++;
		dst+=3;
		x++;
		if (x == w)
		{
			dst += stride;
			x = 0;
		}
	}
}


/**
 * Store a dataset to the destination 32bpp image data set. Each color and alpha channel
 * is set to the same color given by the value in the dataset
 *
 * The parameter scale is multiplied to each value before it's stored
 * \param dst		pointer to destination data storage
 * \param scale		scale parameter for the value before storing to destination data
 * \param stride	stride if the data should be stored to a pixbuf
 */
template<>
void inline CData<float>::storeData32bppAndScale(unsigned char *dst, float scale, size_t stride)
{
	if (stride != 0)
		stride -= w*4;	// just compute the last offset

	float *src = data;
	int x = 0;
	for (size_t i = 0; i < pixel_size; i++)
	{
		unsigned char value = (unsigned char)((*src)*scale);
		*(dst+0) = value;
		*(dst+1) = value;
		*(dst+2) = value;
		*(dst+3) = 0;

		src++;
		dst+=4;
		x++;
		if (x == w)
		{
			dst += stride;
			x = 0;
		}
	}
}



/**
 * Store a dataset to the destination 24bpp image data set. Each color channel
 * is set to the same color given by the value in the dataset
 *
 * The parameters min and max specify the output window where the values are lineary
 * transformed to before storing to the destination dataset
 * \param dst		pointer to destination data storage
 * \param min		minimum value e. g. for contrast enhancement
 * \param max		maximum value e. g. for contrast enhancement
 * \param stride	stride if the data should be stored to a pixbuf
 */
template<>
void inline CData<float>::storeData24bppAndScale(unsigned char *dst, float min, float max, size_t stride)
{
	if (stride != 0)
		stride -= w*3;	// just compute the last offset

	float scale = 255.0f/(max - min);
	float *src = data;
	int x = 0;
	for (size_t i = 0; i < pixel_size; i++)
	{
		unsigned char value = (unsigned char)(((*src)-min)*scale);
		*(dst++) = value;
		*(dst++) = value;
		*(dst++) = value;

		src++;
		x++;
		if (x == w)
		{
			dst += stride;
			x = 0;
		}
	}
}


/**
 * Store a slice of the dataset to the destination 24bpp image data set. Each color channel
 * is set to the same color given by the value in the dataset
 *
 * The parameters min and max specify the output window where the values are lineary
 * transformed to before storing to the destination dataset
 * \param dst		pointer to destination data storage
 * \param min		minimum value e. g. for contrast enhancement
 * \param max		maximum value e. g. for contrast enhancement
 * \param stride	stride if the data should be stored to a pixbuf
 * \param slice		id of slide to be stored to 'dst'
 */
template<>
void inline CData<float>::storeData24bppAndScaleSlice(unsigned char *dst, float min, float max, int slice, size_t stride)
{
	if (stride != 0)
		stride -= w*3;	// just compute the last offset

	float scale = 255.0f/(max - min);
	float *src = data + w*h*slice;
	int x = 0;
	for (int i = 0; i < w*h; i++)
	{
		unsigned char value = (unsigned char)(((*src)-min)*scale);
		*(dst++) = value;
		*(dst++) = value;
		*(dst++) = value;

		src++;
		x++;
		if (x == w)
		{
			dst += stride;
			x = 0;
		}
	}
}


/**
 * Store a dataset to the destination 32bpp image data set. Each color and alpha channel
 * is set to the same color given by the value in the dataset
 *
 * The parameters min and max specify the output window where the values are lineary
 * transformed to before storing to the destination dataset
 * \param dst		pointer to destination data storage
 * \param min		minimum value e. g. for contrast enhancement
 * \param max		maximum value e. g. for contrast enhancement
 * \param stride	stride if the data should be stored to a pixbuf
 */
template<>
void inline CData<float>::storeData32bppAndScale(unsigned char *dst, float min, float max, size_t stride)
{
	if (stride != 0)
		stride -= w*4;	// just compute the last offset

	float scale = 255.0f/(max - min);
	float *src = data;
	int x = 0;
	for (size_t i = 0; i < pixel_size; i++)
	{
		unsigned char value = (unsigned char)(((*src)-min)*scale);
		*(dst++) = value;
		*(dst++) = value;
		*(dst++) = value;
		*(dst++) = 0;

		src++;
		x++;
		if (x == w)
		{
			dst += stride;
			x = 0;
		}
	}
}


/**
 * Same as the "float datastorage container" except that each colorchannel has it's corresponding value of the vector
 */
template<>
void inline CData<CVector<3,float> >::storeData24bppAndScale(unsigned char *dst, float min, float max, size_t stride)
{
	if (stride != 0)
		stride -= w*3;	// just compute the last offset

	float scale = 255.0f/(max - min);
	CVector<3,float> *src = data;
	int x = 0;
	for (size_t i = 0; i < pixel_size; i++)
	{
		*(dst++) = (unsigned char)((src->data[0]-min)*scale);
		*(dst++) = (unsigned char)((src->data[1]-min)*scale);
		*(dst++) = (unsigned char)((src->data[2]-min)*scale);

		src++;
		x++;
		if (x == w)
		{
			dst += stride;
			x = 0;
		}
	}
}

/**
 * Same as the "float datastorage container" except that each colorchannel has it's corresponding value of the vector
 */
template<>
void inline CData<CVector<3,float> >::storeData24bppAndScaleSlice(unsigned char *dst, float min, float max, int slice, size_t stride)
{
	if (stride != 0)
		stride -= w*3;	// just compute the last offset

	float scale = 255.0f/(max - min);
	CVector<3,float> *src = data + w*h*slice;
	int x = 0;
	for (int i = 0; i < w*h; i++)
	{
		*(dst++) = (unsigned char)((src->data[0]-min)*scale);
		*(dst++) = (unsigned char)((src->data[1]-min)*scale);
		*(dst++) = (unsigned char)((src->data[2]-min)*scale);

		src++;
		x++;
		if (x == w)
		{
			dst += stride;
			x = 0;
		}
	}
}

/**
 * Same as the "float datastorage container" except that each colorchannel has it's corresponding value of the vector
 */
template<>
void inline CData<CVector<3,float> >::storeData32bppAndScale(unsigned char *dst, float min, float max, size_t stride)
{
	if (stride != 0)
		stride -= w*4;	// just compute the last offset

	float scale = 255.0f/(max - min);
	CVector<3,float> *src = data;
	int x = 0;
	for (size_t i = 0; i < pixel_size; i++)
	{
		*(dst++) = (unsigned char)((src->data[0]-min)*scale);
		*(dst++) = (unsigned char)((src->data[1]-min)*scale);
		*(dst++) = (unsigned char)((src->data[2]-min)*scale);
		*(dst++) = 0;

		src++;
		x++;
		if (x == w)
		{
			dst += stride;
			x = 0;
		}
	}
}

/**
 * Same as the "float datastorage container" except that each colorchannel has it's corresponding value of the vector
 */
template<>
void inline CData<CVector<3,float> >::storeData24bppAndScale(unsigned char *dst, float scale, size_t stride)
{
	if (stride != 0)
		stride -= w*3;	// just compute the last offset

	CVector<3,float> *src = data;
	int x = 0;
	for (size_t i = 0; i < pixel_size; i++)
	{
		*(dst+0) = (unsigned char)((*src)[0]*scale);
		*(dst+1) = (unsigned char)((*src)[1]*scale);
		*(dst+2) = (unsigned char)((*src)[2]*scale);

		src++;
		dst+=3;
		x++;
		if (x == w)
		{
			dst += stride;
			x = 0;
		}
	}
}

/**
 * Same as the "float datastorage container" except that each colorchannel has it's corresponding value of the vector
 */
template<>
void inline CData<CVector<3,float> >::storeData32bppAndScale(unsigned char *dst, float scale, size_t stride)
{
	if (stride != 0)
		stride -= w*4;	// just compute the last offset

	CVector<3,float> *src = data;
	int x = 0;
	for (size_t i = 0; i < pixel_size; i++)
	{
		*(dst+0) = (unsigned char)((*src)[0]*scale);
		*(dst+1) = (unsigned char)((*src)[1]*scale);
		*(dst+2) = (unsigned char)((*src)[2]*scale);
		*(dst+3) = 0;

		src++;
		dst+=4;

		x++;
		if (x == w)
		{
			dst += stride;
			x = 0;
		}
	}
}

#endif
