/*
 * This file is part of the segmentation and registration toolkit (SphereSegRegTk).
 *
 * SphereSegRegTk is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * SphereSegRegTk is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with SphereSegRegTk.  If not, see <http://www.gnu.org/licenses/>.
 */
#ifndef FILE_H
#define FILE_H

#include <stdlib.h>
#include <stdio.h>
#include <iostream>


/// File content access
/**
 * Simple file access handler
 *
 * Read out the contents of a file, store the file length to *fileLength and return the pointer to the data
 * if terminateNull is true, the content string is terminated by appending simply a 0 byte
 */
class CFile
{
public:

/**
 * read out the contents of a file
 * \param filepath	path to the file
 * \param fileLength	pointer to integer variable to store the read length of file
 * \param terminateNull	set to true, if the array should be terminated by '0' (useful for textfiles)
 * \return		pointer to file contents (has to be freed by the callee)
 */
static void *fileContents(const char *filepath, int *fileLength, bool terminateNull)
{
	int length;
	char *data;

	FILE *file = fopen(filepath, "rb");
	if (file == NULL)
	{
		perror("fileContent() fopen");
		return 0;
	}

	fseek(file, 0, SEEK_END);
	length = ftell(file);
	fseek(file, 0, SEEK_SET);

	if (terminateNull)	data = new char[length + 1];
	else			data = new char[length];

	if (!data)
	{
		fprintf(stderr, "fileContent() out of memory!");
		return 0;
	}

	int readLength;
	readLength = fread(data, 1, length, file);
	fclose(file);

	if (readLength != length)
		fprintf(stderr, "readLength (%i) != fileLength (%i) (ignoring)\n", readLength, length);

	if (fileLength)
		*fileLength = length;

	if (terminateNull)
		data[length] = 0;

	return data;
}

/**
 * store the data to a file
 * \param filepath	path to the file where the data should be stored
 * \param data		pointer to the data which should be stored
 * \param length	length of data field in bytes
 */
static bool storeToFile(const char *filepath, void *data, int length)
{
	FILE *file = fopen(filepath, "w");
	if (file == NULL)
	{
		perror("fileContent() fopen");
		return 0;
	}

	// TODO: error check
	fwrite(data, length, 1, file);
	fclose(file);

	return true;
}


/**
 * load data from a file
 * \param filepath	path to file to load
 * \param data		storage area where the file contents are stored
 * \param max_length	maximum length
 * \return		-1 on error, otherwise the loaded bytes
 */
static int loadFromFile(const char *filepath, void *data, int max_length)
{
	int length;

	FILE *file = fopen(filepath, "w");
	if (file == NULL)
	{
		perror("fileContent() fopen");
		return 0;
	}

	fseek(file, 0, SEEK_END);
	length = ftell(file);
	fseek(file, 0, SEEK_SET);

	if (max_length != -1)
		if (length > max_length-1)
		{
			std::cerr << "not enough space (filesize: " << length << std::endl;
			return -1;
		}

	// TODO: error check
	int read_length = fread(data, length, 1, file);
	fclose(file);

	return read_length;
}


};

#endif
