/*
 * This file is part of the segmentation and registration toolkit (SphereSegRegTk).
 *
 * SphereSegRegTk is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * SphereSegRegTk is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with SphereSegRegTk.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef CFILTERGRADIENT3_HPP
#define CFILTERGRADIENT3_HPP

#include "CVector.hh"
#include "CData.hpp"
#include "CMemoryContainer.hpp"


/// <b>SegRegTk:</b> Gradient Filter
/**
 * 3D gradient filter
 */
template <typename T>
class CFilterGradient
{
public:
#if 1
	/**
	 * Apply a gradient filter with a postprocessed smoothing filter
	 * \param src	scalar (potential) data array
	 * \param dst	destination vector array where the gradient vectors computed from the scalar field are stored
	 */
	void filter(CData<float> &src, CData<CVector<3,T> > &dst)
	{
		CData<CVector<3,T> > dst_buf;

		dst.resize(src.w, src.h, src.d);
		dst_buf.resize(src.w, src.h, src.d);

		CVector<3,T> vector0(0.0);

		for (int x = 0; x < src.w; x++)
		{
			// y \in {0, src.h-1}
			dst.set(x, 0, 0, vector0);
			dst.set(x, src.h-1, 0, vector0);
			dst.set(x, 0, src.d-1, vector0);
			dst.set(x, src.h-1, src.d-1, vector0);
			dst_buf.set(x, 0, 0, vector0);
			dst_buf.set(x, src.h-1, 0, vector0);
			dst_buf.set(x, 0, src.d-1, vector0);
			dst_buf.set(x, src.h-1, src.d-1, vector0);
		}

		for (int y = 0; y < src.h; y++)
		{
			// x \in {0, src.w-1}
			dst.set(0, y, 0, vector0);
			dst.set(src.w-1, y, 0, vector0);
			dst.set(0, y, src.d-1, vector0);
			dst.set(src.w-1, y, src.d-1, vector0);
			dst_buf.set(0, y, 0, vector0);
			dst_buf.set(src.w-1, y, 0, vector0);
			dst_buf.set(0, y, src.d-1, vector0);
			dst_buf.set(src.w-1, y, src.d-1, vector0);
		}

		for (int z = 0; z < src.d; z++)
		{
			// x \in {0, src.w-1}
			dst.set(0, 0, z, vector0);
			dst.set(src.w-1, 0, z, vector0);
			dst.set(0, src.h-1, z, vector0);
			dst.set(src.w-1, src.h-1, z, vector0);
			dst_buf.set(0, 0, z, vector0);
			dst_buf.set(src.w-1, 0, z, vector0);
			dst_buf.set(0, src.h-1, z, vector0);
			dst_buf.set(src.w-1, src.h-1, z, vector0);
		}

		int percent = 0;
		for (int z = 1; z < src.d-1; z++)
		{
			for (int y = 1; y < src.h-1; y++)
			{
				for (int x = 1; x < src.w-1; x++)
				{
					dst_buf.set(x, y, z, CVector<3,float>(
							- src.get(x-1, y, z)
							+ src.get(x+1, y, z)
								,
							- src.get(x, y-1, z)
							+ src.get(x, y+1, z)
								,
							- src.get(x, y, z-1)
							+ src.get(x, y, z+1)
						));
				}
			}

			if (100*(z-1)/(src.d-2) > percent)
			{
				std::cout << "." << std::flush;
				percent++;
			}
		}

		std::cout << "|" << std::flush;

		/*
		 * smooth the gradients for better particle tracing
		 * this creates a much better result for the particle tracer
		 */
		for (int s = 0; s < 1; s++)
		{
			percent = 0;
			for (int z = 1; z < src.d-1; z++)
			{
				for (int y = 1; y < src.h-1; y++)
				{
					for (int x = 1; x < src.w-1; x++)
					{
						CVector<3,T> gradient =
								(
									dst_buf.get(x-1, y-1, z-1)	+
									dst_buf.get(x, y-1, z-1)	+
									dst_buf.get(x+1, y-1, z-1)	+

									dst_buf.get(x-1, y, z-1)	+
									dst_buf.get(x, y, z-1)		+
									dst_buf.get(x+1, y, z-1)	+

									dst_buf.get(x-1, y+1, z-1)	+
									dst_buf.get(x, y+1, z-1)	+
									dst_buf.get(x+1, y+1, z-1)	+

									dst_buf.get(x-1, y-1, z)	+
									dst_buf.get(x, y-1, z)		+
									dst_buf.get(x+1, y-1, z)	+

									dst_buf.get(x-1, y, z)		+
									dst_buf.get(x, y, z)		+
									dst_buf.get(x+1, y, z)		+

									dst_buf.get(x-1, y+1, z)	+
									dst_buf.get(x, y+1, z)		+
									dst_buf.get(x+1, y+1, z)	+

									dst_buf.get(x-1, y-1, z+1)	+
									dst_buf.get(x, y-1, z+1)	+
									dst_buf.get(x+1, y-1, z+1)	+

									dst_buf.get(x-1, y, z+1)	+
									dst_buf.get(x, y, z+1)		+
									dst_buf.get(x+1, y, z+1)	+

									dst_buf.get(x-1, y+1, z+1)	+
									dst_buf.get(x, y+1, z+1)	+
									dst_buf.get(x+1, y+1, z+1)
								);// * (T)(1.0/27.0);	// normalize result!

						dst.set(x, y, z, gradient);
					}
				}

				if (100*(z-1)/(dst.d-2) > percent)
				{
					std::cout << "." << std::flush;
					percent++;
				}
			}
		}
	}
#else

	void filter(CData<float> &src, CData<CVector<3,T> > &dst)
	{
		dst.resize(src.w, src.h, src.d);

		CVector<3,T> vector0(0.0);

		for (int x = 0; x < src.w; x++)
		{
			// y \in {0, src.h-1}
			dst.set(x, 0, 0, vector0);
			dst.set(x, src.h-1, 0, vector0);
			dst.set(x, 0, src.d-1, vector0);
			dst.set(x, src.h-1, src.d-1, vector0);
		}

		for (int y = 0; y < src.h; y++)
		{
			// x \in {0, src.w-1}
			dst.set(0, y, 0, vector0);
			dst.set(src.w-1, y, 0, vector0);
			dst.set(0, y, src.d-1, vector0);
			dst.set(src.w-1, y, src.d-1, vector0);
		}

		for (int z = 0; z < src.d; z++)
		{
			// x \in {0, src.w-1}
			dst.set(0, 0, z, vector0);
			dst.set(src.w-1, 0, z, vector0);
			dst.set(0, src.h-1, z, vector0);
			dst.set(src.w-1, src.h-1, z, vector0);
		}

		int percent = 0;
		for (int z = 1; z < src.d-1; z++)
		{
			for (int y = 1; y < src.h-1; y++)
			{
				for (int x = 1; x < src.w-1; x++)
				{
					dst.set(x, y, z, CVector<3,float>(
							- src.get(x-1, y, z)
							+ src.get(x+1, y, z)
								,
							- src.get(x, y-1, z)
							+ src.get(x, y+1, z)
								,
							- src.get(x, y, z-1)
							+ src.get(x, y, z+1)
						));
				}
			}

			if (100*(z-1)/(src.d-2) > percent)
			{
				std::cout << "." << std::flush;
				percent++;
			}
		}
	}
#endif

	/**
	 * Copy the contents of the gradient array to a memory container
	 */
	static void save(CMemoryContainer &container, CData<CVector<3,T> > &data)
	{
		container.push((int)data.w);
		container.push((int)data.h);
		container.push((int)data.d);
		for (size_t i = 0; i < data.pixel_size; i++)
			container.push3(data.data[i].data);
	}

	/**
	 * Load the contents from a memory container to the gradient array
	 */
	static bool load(CMemoryContainer &container, CData<CVector<3,T> > &data)
	{
		int w,h,d;
		container.get(&w);
		container.get(&h);
		container.get(&d);

		data.resize(w, h, d);

		for (size_t i = 0; i < data.pixel_size; i++)
			container.get3(data.data[i].data);

		return true;
	}
};

#endif
