/*
 * This file is part of the segmentation and registration toolkit (SphereSegRegTk).
 *
 * SphereSegRegTk is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * SphereSegRegTk is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with SphereSegRegTk.  If not, see <http://www.gnu.org/licenses/>.
 */
/**
 * this filter builds transformation matrices using graph nodes of both graphs to
 * transform one graph onto another, compute the distance between both graphs and use the matrix
 * with the lowest distance as the best matrix
 */

#ifndef __CFILTERMATCHINGSIMPLE3_HPP
#define __CFILTERMATCHINGSIMPLE3_HPP

#include <kdtree++/kdtree.hpp>
#include "CMatrix4.hh"
#include "CMatrix3.hh"
#include "CPointRadius.hh"
#include <pthread.h>



/// <b>SegRegTk:</b> Matching filter (affine transformation)
template< typename T >
class CFilterMatchingSimple
{
private:

	/**
	 * class to store the testing points used to create the transformation matrix
	 */
	class CMatchPoint
	{
	public:
		CPointRadius<3,T> point;
		typename CGraph<3,T>::CEdge *edge;	// matching edge
		typename list<typename CGraph<3,T>::CNode>::iterator iter;	// iterator on edge
		bool used;				// node is used to compute matching

		CMatchPoint(	CVector<3,T> p_pos,
				T p_radius,
			       	typename CGraph<3,T>::CEdge &p_edge,
			       	typename list<typename CGraph<3,T>::CNode>::iterator &p_iter
		)
		{
			point.pos = p_pos;
			point.radius = p_radius;	// not used so far
			edge = &p_edge;
			used = false;
			iter = p_iter;
		}

		CMatchPoint &operator=(const CMatchPoint &m)
		{
			point = m.point;
			edge = m.edge;
			used = m.used;
			iter = m.iter;
		}

		static bool cmp_sorted_points(CMatchPoint &p1, CMatchPoint &p2)
		{
			//return p1.point.radius < p2.point.radius;
			return p1.point.radius > p2.point.radius;
		}

	};

	inline static T tree_access_control(CMatchPoint* p, int k)
	{
	       	return p->point.pos.data[k];
	}

	typedef KDTree::KDTree<	3,		// dimension
				CMatchPoint*,	// storage data type
				std::pointer_to_binary_function<CMatchPoint*,int,T>	// access function to data in storage type { float (CPoint*)[int] }
			> kdtree_type;

	kdtree_type *tree1;	///< kdtree handler for nodes of graph1
	kdtree_type *tree2;	///< kdtree handler for nodes of graph2



private:
	// constraints for matrix values
	T max_untransformed_distance;		///< CONSTRAINT: Maximum distance between untransformed nodes used for mapping matrix creation
	T max_untransformed_distance_pow_2;
	T max_rotation_angle;			///< MATRIX CONSTRAINT: Maximum rotation angle
	T min_scale_factor;			///< MATRIX CONSTRAINT: Minimum scale factor
	T max_scale_factor;			///< MATRIX CONSTRAINT: Maximum scale factor
	T min_shear_factor;			///< MATRIX CONSTRAINT: Minimum shear component
	T max_shear_factor;			///< MATRIX CONSTRAINT: Maximum shear component
	CVector<3,T> max_translation_dist;	///< MATRIX CONSTRAINT: Maximum translation

	CVector<3,T> pre_translation;		///< pretranslation components
	T min_matrix_generation_point_distance;	///< minimum distance between the points of one graph
	T min_matrix_generation_point_distance_pow_2;

	bool use_mutual_graph_distance;		///< true, if the matching distance is computed matching graph1 to graph2 and graph2 to graph1
	T adaptive_max_points;			///< ADAPTIVE DISTANCE: Maximum points restricting the point distance
	T adaptive_start_point_distance;	///< ADAPTIVE DISTANCE: start point distance to start adaptive measurement
	T adaptive_start_point_distance_pow_2;
	T adaptive_start_radius;		///< ADAPTIVE DISTANCE: start radius (works not properly, set to -1 to disable)

	int number_of_threads;			///< number of threads to use for matching

	bool use_point_to_line_distance;	///< use point to line distance for graph distance computations

public:
	mat4f best_matrix_trans;		///< transformation matrix with minimum distance
	mat4f best_pre_translated_matrix_trans;	///< transformation matrix with minimum distance and applied pretranslation to center image
	T min_distance;				///< minimum computed distance between graph1 and graph2

	void setParameters(
			T p_max_untransformed_distance = 100,		///< CONSTRAINT: Maximum distance between untransformed nodes used for mapping matrix creation
			T p_min_matrix_generation_point_distance = 100,	///< minimum distance between the points of one graph
			bool p_use_mutual_graph_distance = false,	///< true, if the matching distance is computed matching graph1 to graph2 and graph2 to graph1

			T p_adaptive_max_points = 0,			///< ADAPTIVE DISTANCE: Maximum points restricting the point distance
			T p_adaptive_start_point_distance = 512,	///< ADAPTIVE DISTANCE: start point distance to start adaptive measurement
			T p_adaptive_start_radius = 50,			///< ADAPTIVE DISTANCE: start radius (works not properly, set to -1 to disable)

			bool p_use_point_to_line_distance = true,	///< use point to line distance for graph distance computations
			int p_number_of_threads = 1,			///< number of threads to use for matching

			// matrix restrictions
			T p_max_rotation_angle = 20,		///< MATRIX CONSTRAINT: Maximum rotation angle

			T p_min_scale_factor = 0.7,		///< MATRIX CONSTRAINT: Minimum scale factor
			T p_max_scale_factor = 1.3,		///< MATRIX CONSTRAINT: Maximum scale factor
			T p_min_shear_factor = 0.7,		///< MATRIX CONSTRAINT: Minimum shear component
			T p_max_shear_factor = 1.3,		///< MATRIX CONSTRAINT: Maximum shear component

			T p_max_translation_dist_x = 100,	///< MATRIX CONSTRAINT: Maximum translation
			T p_max_translation_dist_y = 100,
			T p_max_translation_dist_z = 100,

			T p_pre_translation_x = 256,		///< pretranslation components to center object
			T p_pre_translation_y = 256,
			T p_pre_translation_z = 256
	)
	{
		max_untransformed_distance	= p_max_untransformed_distance;
		max_untransformed_distance_pow_2	= max_untransformed_distance*max_untransformed_distance;

		min_matrix_generation_point_distance = p_min_matrix_generation_point_distance;
		min_matrix_generation_point_distance_pow_2 = min_matrix_generation_point_distance*min_matrix_generation_point_distance;

		use_mutual_graph_distance = p_use_mutual_graph_distance;

		adaptive_max_points = p_adaptive_max_points;
		adaptive_start_point_distance = p_adaptive_start_point_distance;
		adaptive_start_point_distance_pow_2 = p_adaptive_start_point_distance*p_adaptive_start_point_distance;;
		adaptive_start_radius = p_adaptive_start_radius;

		use_point_to_line_distance = p_use_point_to_line_distance;

		number_of_threads = p_number_of_threads;

		max_rotation_angle	= p_max_rotation_angle;
		min_scale_factor	= p_min_scale_factor;
		max_scale_factor	= p_max_scale_factor;
		min_shear_factor	= p_min_shear_factor;
		max_shear_factor	= p_max_shear_factor;

		max_translation_dist[0]	= p_max_translation_dist_x;
		max_translation_dist[1]	= p_max_translation_dist_y;
		max_translation_dist[2]	= p_max_translation_dist_z;

		pre_translation[0] = p_pre_translation_x;
		pre_translation[1] = p_pre_translation_y;
		pre_translation[2] = p_pre_translation_z;
	}


CFilterMatchingSimple()
{
	setParameters();

	tree1 = new kdtree_type(std::ptr_fun(tree_access_control));
	tree2 = new kdtree_type(std::ptr_fun(tree_access_control));

	// nothing to do
#if 0
	// check matrix solving
	mat4f m;
	vec4f b;
	m[0][0] = -138;	m[0][1] = 124;	m[0][2] = -342;		m[0][3] = 183;		b[0] = 45;
	m[1][0] = 432;	m[1][1] = 922;	m[1][2] = 554;		m[1][3] = 13;		b[1] = 26;
	m[2][0] = 472;	m[2][1] = -772;	m[2][2] = -146;		m[2][3] = 947;		b[2] = 12;
	m[3][0] = 174;	m[3][1] = 742;	m[3][2] = 346;		m[3][3] = 47;		b[3] = 12;

	std::cout << std::endl;
	std::cout << std::endl;

	std::cout << "Matrix:" << std::endl;
	std::cout << m << std::endl;
	std::cout << std::endl;
	std::cout << "Vector b: " << b << std::endl;
	std::cout << std::endl;
	std::cout << "solving" << std::endl;
	vec4f s = m.solve(b);
	std::cout << std::endl;
	std::cout << "solution of solve(M,b) => " << s << std::endl;
	std::cout << endl;
	std::cout << "multiplying matrix with solution" << std::endl;
	vec4f a = m*s;
	std::cout << "M*s = " << a << std::endl;
	vec4f x = a-b;
	std::cout << "M*s-b = " << x << "    length: " << x.length() << std::endl;


	std::cout << std::endl;
	std::cout << std::endl;
#endif
}



#define MATCHING_FILTER_SIMPLE_ABORT_ON_DECOMPOSITION	1
/**
 * decompose a transformation matrix into
 *
 * - rotation around x	Rx
 *
 * - rotation around y	Ry
 *
 * - rotation around z	Rz
 *
 * - scaling		S
 *
 * - shearing x, y, z	H
 *
 * - translation	T
 *
 * M = H * S * Rz * Rx * Ry * T
 *
 * the values are stored to the references except matrix_orig
 */
inline bool decompose_and_filter_matrix(
		CMatrix4<T> &matrix_orig,
		CVector<3,T> &rotation_angles,	// rotation angles around z, x and y axis
		CVector<3,T> &scaling,		// scaling components (diagonal)
		CVector<3,T> &shearing,		// (values above the diagonals)
		CVector<3,T> &translation	// translation
)
{

	CMatrix4<T> matrix_decomp;	// matrix for decomposition

	matrix_decomp = matrix_orig;

	/*
	 * compute translation
	 *
	 * M = M' * T

	 * setting m'[n,3] to 0 gives the following equations:
	 * ( m[0,3] )   ( m[0,0]  m[0,1]  m[0,2] )   ( t[0] )
	 * ( m[1,3] ) = ( m[1,0]  m[1,1]  m[1,2] ) * ( t[1] )
	 * ( m[2,3] )   ( m[2,0]  m[2,1]  m[2,2] )   ( t[2] )
	 *
	 * solve for variable t gives the solution
	 */
	CMatrix3<T> matrix_tmp;
	matrix_tmp[0][0] = matrix_decomp[0][0];
	matrix_tmp[0][1] = matrix_decomp[0][1];
	matrix_tmp[0][2] = matrix_decomp[0][2];
	matrix_tmp[1][0] = matrix_decomp[1][0];
	matrix_tmp[1][1] = matrix_decomp[1][1];
	matrix_tmp[1][2] = matrix_decomp[1][2];
	matrix_tmp[2][0] = matrix_decomp[2][0];
	matrix_tmp[2][1] = matrix_decomp[2][1];
	matrix_tmp[2][2] = matrix_decomp[2][2];
	CVector<3,T> left(matrix_decomp[0][3], matrix_decomp[1][3], matrix_decomp[2][3]);

	translation = matrix_tmp.solve(left);

#if MATCHING_FILTER_SIMPLE_ABORT_ON_DECOMPOSITION
	if (CMath<T>::abs(translation[0]) > max_translation_dist[0])	return false;
	if (CMath<T>::abs(translation[1]) > max_translation_dist[1])	return false;
	if (CMath<T>::abs(translation[2]) > max_translation_dist[2])	return false;
#endif

	matrix_decomp[0][3] = 0;
	matrix_decomp[1][3] = 0;
	matrix_decomp[2][3] = 0;

	/*
	 * analyze if the transformation matrix is valid
	 *
	 * QR decomposition is used
	 * the original matrix [matrix] is labeled as M
	 *
	 * the 2d rotation matrix has the form
	 *
	 *     (  cos   sin)
	 * R = (           )
	 *     ( -sin   cos)
	 *
	 * Ry as the rotation around the y axis (y remains the same) could be represented as
	 *
	 *      (  cos(a)   0    sin(a)   0 )
	 * Ry = (   0       1      0      0 )
	 *      ( -sin(a)   0    cos(a)   0 )
	 *      (   0       0      0      1 )
	 *
	 * the decomposition computes the rotational matrices Rx,y,z so that it holds that
	 * Rn is the rotation around the axis n
	 *
	 * the rotational decomposition is then specified by the following matrixes
	 * P' = D1 * Rz * Rx * Ry * P
	 *
	 *-------------------------
	 * SETTING M[2,0] TO 0 with rotation around y axis
	 *-------------------------
	 * decomposing the initial matrix in the form
	 *
	 * M = T1 * Ry          where T1[2,0] = 0 and T1[3] = (0,0,0,1)
	 * or
	 * M * Ry^(-1) = T1     where Ry^(-1) is the rotation with the negative angle
	 *
	 * T1[2,0] gives the constraint:
	 *
	 * M[2,0] * cos(ay') - M[2,2] * sin(ay') = 0
	 * <=> tan(ay') = M[2,0] / M[2,2]
	 *
	 * setting M[2,0] to 0 is done by multiplying M with the inverse rotation matrix
	 * specified by the angle ay'
	 * 
	 *-------------------------
	 * SETTING M[2,1] TO 0 with rotation around x axis
	 *-------------------------
	 * M = T1 * Ry = T2 * Rx * Ry         where T2[2,0] = 0, T2[2,1] = 0 and T2[3] = (0,0,0,1)
	 * or
	 * M * Rx^(-1) = T2
	 *
	 * M[2,1] * cos(ax') - M[2,2] * sin(ax') = 0
	 * <=> tan(ax') = M[2,1] / M[2,2]
	 *
	 * this rotation matrix does not change the value T1[2,0]==T2[2,0] which is 0
	 * 
	 *-------------------------
	 * SETTING M[1,0] TO 0
	 *-------------------------
	 * M = T2 * Rx * Ry = T3 * Rz * Rx * Ry         where T3[2,0] = T3[2,1] = T3[1,0] = 0 and T3[3] = (0,0,0,1)
	 * or
	 * M * Rz^(-1) = T3
	 *
	 * M[1,0] * cos(az') - M[1,1] * sin(az') = 0
	 * <=> tan(az') = M[1,0] / M[1,1]
	 *
	 */

	CMatrix4<T> matrix_rot;		// matrix for rotations
	/**************************
	 * Y ROTATION MATRIX
	 */

	if (CMath<T>::abs(matrix_decomp[2][0]) < 0.0001)
	{
		// assum 0 angle for rotation around y axis
		rotation_angles[1] = 0;
	}
	else
	{
		rotation_angles[1] = atan(matrix_decomp[2][0]/matrix_decomp[2][2]);

#if MATCHING_FILTER_SIMPLE_ABORT_ON_DECOMPOSITION
		if (CMath<T>::abs(rotation_angles[1]) > max_rotation_angle)
			return false;
#endif

		// create inverse rotation matrix
		// a is always positive or 0 because it's created by power of 2 scalars
		matrix_rot.loadIdentity();	// initialize rotation matrix with identity matrix
		matrix_rot[0][0] = cos(rotation_angles[1]);
		matrix_rot[0][2] = sin(rotation_angles[1]);
		matrix_rot[2][0] = -sin(rotation_angles[1]);
		matrix_rot[2][2] = cos(rotation_angles[1]);

		matrix_decomp = matrix_decomp * matrix_rot;

		if (CMath<T>::abs(matrix_decomp[2][0]) > 0.001)
		{
#if FILTER_MATCHING_SIMPLE_DEBUG
			std::cout << "[===== DECOMP Y FAILURE ====]" << std::endl;
			std::cout << "rotation angle: " << rotation_angles[1] << std::endl;
			std::cout << "---------------------------" << std::endl;
			std::cout << "rotation matrix" << std::endl;
			std::cout << matrix_rot;
			std::cout << "decomposed matrix" << std::endl;
			std::cout << matrix_decomp;
			std::cout << "[=========================]" << std::endl;
#endif
#if MATCHING_FILTER_SIMPLE_ABORT_ON_DECOMPOSITION
			return false;
#endif
		}
	}

	/**************************
	 * X ROTATION MATRIX
	 */

	if (CMath<T>::abs(matrix_decomp[2][1]) < 0.00001)
	{
		// assum 0 angle for rotation around x axis
		rotation_angles[0] = 0;
	}
	else
	{
		rotation_angles[0] = atan(matrix_decomp[2][1]/matrix_decomp[2][2]);

#if MATCHING_FILTER_SIMPLE_ABORT_ON_DECOMPOSITION
		if (CMath<T>::abs(rotation_angles[0]) > max_rotation_angle)
			return false;
#endif

		// create inverse rotation matrix
		// a is always positive or 0 because it's created by power of 2 scalars
		matrix_rot.loadIdentity();	// initialize rotation matrix with identity matrix
		matrix_rot[1][1] = cos(rotation_angles[0]);
		matrix_rot[1][2] = sin(rotation_angles[0]);
		matrix_rot[2][1] = -sin(rotation_angles[0]);
		matrix_rot[2][2] = cos(rotation_angles[0]);

		matrix_decomp = matrix_decomp * matrix_rot;

		if (	CMath<T>::abs(matrix_decomp[2][0]) > 0.001 ||
			CMath<T>::abs(matrix_decomp[2][1]) > 0.001
		)
		{
#if FILTER_MATCHING_SIMPLE_DEBUG
			std::cout << "[===== DECOMP X FAILURE ====]" << std::endl;
			std::cout << "rotation angle: " << rotation_angles[0] << std::endl;
			std::cout << "---------------------------" << std::endl;
			std::cout << "rotation matrix" << std::endl;
			std::cout << matrix_rot;
			std::cout << "decomposed matrix" << std::endl;
			std::cout << matrix_decomp;
			std::cout << "[=========================]" << std::endl;
#endif
#if MATCHING_FILTER_SIMPLE_ABORT_ON_DECOMPOSITION
			return false;
#endif
		}
	}

	/**************************
	 * Z ROTATION MATRIX
	 */

	if (CMath<T>::abs(matrix_decomp[2][1]) < 0.00001)
	{
		// assum 0 angle for rotation around z axis
		rotation_angles[2] = 0;
	}
	else
	{
		rotation_angles[2] = atan(matrix_decomp[1][0]/matrix_decomp[1][1]);

#if MATCHING_FILTER_SIMPLE_ABORT_ON_DECOMPOSITION
		if (CMath<T>::abs(rotation_angles[2]) > max_rotation_angle)
			return false;
#endif

		// create inverse rotation matrix
		// a is always positive or 0 because it's created by power of 2 scalars
		matrix_rot.loadIdentity();	// initialize rotation matrix with identity matrix
		matrix_rot[0][0] = cos(rotation_angles[2]);
		matrix_rot[0][1] = sin(rotation_angles[2]);
		matrix_rot[1][0] = -sin(rotation_angles[2]);
		matrix_rot[1][1] = cos(rotation_angles[2]);

		matrix_decomp = matrix_decomp * matrix_rot;

		if (	CMath<T>::abs(matrix_decomp[2][0]) > 0.001	||
			CMath<T>::abs(matrix_decomp[2][1]) > 0.001	||
			CMath<T>::abs(matrix_decomp[1][0]) > 0.001
		)
		{
#if FILTER_MATCHING_SIMPLE_DEBUG
			std::cout << "[===== DECOMP Z FAILURE ====]" << std::endl;
			std::cout << "rotation angle: " << rotation_angles[2] << std::endl;
			std::cout << "---------------------------" << std::endl;
			std::cout << "rotation matrix" << std::endl;
			std::cout << matrix_rot;
			std::cout << "decomposed matrix" << std::endl;
			std::cout << matrix_decomp;
			std::cout << "[=========================]" << std::endl;
#endif
#if MATCHING_FILTER_SIMPLE_ABORT_ON_DECOMPOSITION
			return false;
#endif
		}
	}
	/*
	 * where we are now:
	 *
	 * the rotation angles are within the given maximum rotation angles
	 * the remaining matrix is zero below the diagonal and non-zero on and over the diagonal
	 *
	 * R = Rz * Rx * Ry
	 *
	 * we decompose T3 in the shearing and scaling matrices. the scaling is done at last (leftmost matrix)
	 *
	 * T3 = H * S'
	 * 
	 * where the values of S are 0 except on the diagonal who are responsible for the final scaling
	 *
	 *      ( Sx  0   0   0 )
	 * S' = ( 0   Sy  0   0 )
	 *      ( 0   0   Sz  0 )
	 *      ( 0   0   0   1 )
	 *
	 * H is constructed by a matrix with a 1 diagonal and zero below the diagonal
	 *
	 *     ( 1   H1  H2  0 )
	 * H = ( 0   1   H3  0 )
	 *     ( 0   0   1   0 )
	 *     ( 0   0   0   1 )
	 *
	 *        ( Sx  H1*Sy  H2*Sz  0 )
	 * H*S' = ( 0   Sy     H3*Sz  0 )
	 *        ( 0   0      Sz     0 )
	 *        ( 0   0      0      1 )
	 *
	 */

	scaling[0] = matrix_decomp[0][0];
#if MATCHING_FILTER_SIMPLE_ABORT_ON_DECOMPOSITION
	if (scaling[0] < min_scale_factor)	return false;
	if (scaling[0] > max_scale_factor)	return false;
#endif

	scaling[1] = matrix_decomp[1][1];
#if MATCHING_FILTER_SIMPLE_ABORT_ON_DECOMPOSITION
	if (scaling[1] < min_scale_factor)	return false;
	if (scaling[1] > max_scale_factor)	return false;
#endif

	scaling[2] = matrix_decomp[2][2];
#if MATCHING_FILTER_SIMPLE_ABORT_ON_DECOMPOSITION
	if (scaling[2] < min_scale_factor)	return false;
	if (scaling[2] > max_scale_factor)	return false;
#endif

	shearing[0] = matrix_decomp[0][1]/matrix_decomp[1][1];
#if MATCHING_FILTER_SIMPLE_ABORT_ON_DECOMPOSITION
	if (shearing[0] < min_shear_factor)	return false;
	if (shearing[0] > max_shear_factor)	return false;
#endif

	shearing[1] = matrix_decomp[0][2]/matrix_decomp[2][2];
#if MATCHING_FILTER_SIMPLE_ABORT_ON_DECOMPOSITION
	if (shearing[1] < min_shear_factor)	return false;
	if (shearing[1] > max_shear_factor)	return false;
#endif

	shearing[2] = matrix_decomp[1][2]/matrix_decomp[2][2];
#if MATCHING_FILTER_SIMPLE_ABORT_ON_DECOMPOSITION
	if (shearing[2] < min_shear_factor)	return false;
	if (shearing[2] > max_shear_factor)	return false;
#endif

	return true;
}


private:
/**
 * compute the transformation matrix for the given points and return true if the matrix is valid with the given parameters
 * the result is stored to matrix_trans
 */
inline bool compute_matrix(	CMatrix4<T> &matrix_trans,
				CMatchPoint &p1, CMatchPoint &p2, CMatchPoint &p3, CMatchPoint &p4,
				CMatchPoint &o1, CMatchPoint &o2, CMatchPoint &o3, CMatchPoint &o4	)
{
	mat4f matrix;
	CVector<4,T> b;
	CVector<4,T> f;

	mat4f matrix_rot;	// rotation matrix for decomposition
	mat4f matrix_decomp;	// rotational decomposed matrix

	// create matrix
	matrix[0] = p1.point.pos;
	matrix[1] = p2.point.pos;
	matrix[2] = p3.point.pos;
	matrix[3] = p4.point.pos;

	b = CVector<4,T>(o1.point.pos[0], o2.point.pos[0], o3.point.pos[0], o4.point.pos[0]);
	matrix_trans[0] = matrix.solve(b);
	if (isnan(matrix_trans[0][0]))	return false;

	b = CVector<4,T>(o1.point.pos[1], o2.point.pos[1], o3.point.pos[1], o4.point.pos[1]);
	matrix_trans[1] = matrix.solve(b);
	if (isnan(matrix_trans[1][0]))	return false;

	b = CVector<4,T>(o1.point.pos[2], o2.point.pos[2], o3.point.pos[2], o4.point.pos[2]);
	matrix_trans[2] = matrix.solve(b);
	if (isnan(matrix_trans[2][0]))	return false;

	matrix_trans[3] = CVector<4,T>(0, 0, 0, 1);

#if FILTER_MATCHING_SIMPLE_DEBUG
	vec4f check;
	float length;

	length = (o1.point.pos - matrix_trans*p1.point.pos).length();
	if (length > 0.9)
	{
		std::cerr << "matrix element elimination not successful a " << length << std::endl;
		CVector<3,T> mp = matrix_trans*p1.point.pos;
		std::cerr << o1.point.pos << " " << mp << " " << p1.point.pos << std::endl;
		std::cout << std::endl;
		std::cerr << matrix_trans << std::endl;
		return false;
	}

	length = (o2.point.pos - matrix_trans*p2.point.pos).length();
	if (length > 0.9)
	{
		std::cerr << "matrix element elimination not successful b " << length << std::endl;
		std::cerr << matrix_trans << std::endl;
		return false;
	}

	length = (o3.point.pos - matrix_trans*p3.point.pos).length();
	if (length > 0.9)	
	{
		std::cerr << "matrix element elimination not successful c " << length << std::endl;
		std::cerr << matrix_trans << std::endl;
		return false;
	}

	length = (o4.point.pos - matrix_trans*p4.point.pos).length();
	if (length > 0.9)	
	{
		std::cerr << "matrix element elimination not successful d " << length << std::endl;
		std::cerr << matrix_trans << std::endl;
		return false;
	}
#endif

	return true;
}


/**
 * compute the minimum distance of points from graph1 (matchPoints1) to points of 
 * graph2 (matchPoints2) after applying the matrix_trans to the points of graph1 (matchPoints1)
 *
 * the set matchPoints1 is transformed by matrix matrix_trans
 * then the nearest nodes are searched using tree_2
 */
//#define DIST_POSTPROCESSING(dist)	(dist)
//inline T dist_postprocessing(T dist)	{	return CMath<T>::sqrt(dist);	}
inline T dist_postprocessing(T dist)	{	return dist*dist;	}


T compute_minimum_distance(
			CMatrix4<T> &matrix_trans,		// matrix to transform points from graph1
			list< CMatchPoint > &matchPoints1,	// matching points for graph1
			kdtree_type *tree_2,			// kd tree for points of graph2
			T recent_min_distance,
			T scale_factor
)
{
	T distance = 0;

	for (	typename list<CMatchPoint>::iterator iter1 = matchPoints1.begin();
		iter1 != matchPoints1.end(); 
		iter1++
	)
	{
		if (distance > recent_min_distance)	// early termination because this matching is worser than the previously one
			return distance;

		/**
		 * this point directly matches to another point because
		 * matrix is constructed with this point
		 * (deactivated)
		 */
//		if ((*iter1).used)	continue;

		CVector<3,T> &pos1 = (*iter1).point.pos;
		CVector<3,T> trans_pos1 = matrix_trans*pos1;

		CMatchPoint &nearest_point = **(tree_2->find_nearest(&(*iter1), std::numeric_limits<T>::max()).first);

		T min_distance;

		if (!use_point_to_line_distance)
		{
			// use the simple point-2-point distance
			min_distance = nearest_point.point.pos.dist(trans_pos1);
			distance += dist_postprocessing(min_distance)*scale_factor;
			continue;
		}

		/**
		 * get neighbored points and compute smallest distance between the found nearest point and the
		 * previous or next point
		 */
		CVector<3,T> &l2 = nearest_point.point.pos;

		min_distance = std::numeric_limits<T>::max();

		if (nearest_point.iter != nearest_point.edge->nodes.begin())
		{
			// the nearest node is not the start node of the strip -> check forward neighbored node

			typename list<typename CGraph<3,T>::CNode >::iterator prev_node_iter = nearest_point.iter;
			prev_node_iter--;

			typename CGraph<3,T>::CNode &prev_node = *prev_node_iter;

			/**
			 * line is specified by 2 points: L1(px, py, pz) and L2(qx, qy, qz)
			 * we search the distance to the point V
			 *
			 *  we search the minimum distance from V to point on line L1L2:
			 *
			 *     d = | L1 + t*L1L2 - V |
			 *
			 * <=> dx = px + t*(qx - px) - vx = (px - vx) + t*(qx - px)
			 *     dy = py + t*(qy - py) - vy = (py - vy) + t*(qy - py) = VL1 + t*L1L2
			 *     dz = pz + t*(qz - pz) - vz = (pz - vz) + t*(qz - pz)
			 *
			 * <=> d = sqrt(dx^2 + dy^2 + dz^2)
			 *
			 * minimize => d' = 0
			 *
			 * <=> (VL1^2 + 2*t*(VL1*L1L2) + t^2*L1L2^2)' = 0
			 *
			 * <=> 2*VL1*L1L2 = -2*t*L1L2^2
			 *
			 * <=> t = -((VL1*L1L2))/(L1L2^2)
			 *
			 * R = L1+t*L1L2
			 */
			CVector<3,T> &l1 = prev_node.pos;	// line points
			CVector<3,T> dir = l2 - l1;	// direction is always != 0 because the points have to differ

			T divisor = dir.getDotProd(dir);	// always positive
			if (divisor > 0.0001)
			{
				T t = -(l1-trans_pos1).getDotProd(dir)/divisor;
				if (t >= 0.0 && t <= 1.0)
					min_distance = (l1 + dir*t).dist(trans_pos1);	// TODO: use dist2 for comparison
			}
		}

		typename list<typename CGraph<3,T>::CNode >::iterator next_node_iter = nearest_point.iter;
		next_node_iter++;
		if (next_node_iter != nearest_point.edge->nodes.end())
		{
			typename CGraph<3,T>::CNode &next_node = *next_node_iter;

			CVector<3,T> &l1 = next_node.pos;	// line points
			CVector<3,T> dir = l2 - l1;	// direction is always != 0 because the points have to differ

			T divisor = dir.getDotProd(dir);	// always positive
			if (divisor > 0.0001)
			{
				T t = -(l1-trans_pos1).getDotProd(dir)/divisor;
				if (t >= 0.0 && t <= 1.0)
				{
					T tmp_min_distance = (l1 + dir*t).dist(trans_pos1);
					if (tmp_min_distance < min_distance)
						min_distance = tmp_min_distance;
				}
			}
		}

		if (min_distance == std::numeric_limits<T>::max())
		{
			// if no distance could be computed, use simple distance computation
			// this can e.g. happen, if there's just one node for the whole graph
			min_distance = nearest_point.point.pos.dist(trans_pos1);
			distance += dist_postprocessing(min_distance)*scale_factor;
		}
		else
		{
			distance += dist_postprocessing(min_distance)*scale_factor;
		}
	}

	return distance;
}


void clear_tree1()
{
	tree1->clear();
}
void clear_tree2()
{
	tree2->clear();
}


/**
 * apply filter
 *
 * the filter iterates through all possible combinations of the graph starting and ending nodes to measure the
 * distance between nodes of graph1 compared to nodes to graph1
 *
 * to avoid multiple equal comparsions, the algorithms works as follows:
 * - graph1 is used as graph with less or equal nodes as graph2
 * - the first step is to iterate over matching the first node of graph1 to all possible
 *   nodes of graph2 while calling the recursion again. this recursive step can be seen as removing
 *   the node from graph1 and also the node to match from graph2.
 * - the resursive step itself handles again 2 sets without the first node of graph1 and the matching
 *   node of graph2.
 * - note: the ith node of graph1 is also handled in the ith recursion
 *
 * TODO:
 *
 * A)
 * change creation of matrix with p_a, p_b, p_c with the creation of all possible matchings
 * this avoids duplicate computations of the matrix p_a, p_b and p_c
 * well - this is not necessary for this simple filter...
 *
 *
 * B)
 * create more points for edges - dont use just middle point
 * restrict creation by amount of overall points for graphs and minimum distance between
 * points on edge
 */

void filter_helper_get_adaptive_values(list< CMatchPoint > &matchPoints, T &min_mg_dist, T &min_radius)
{
	min_radius = adaptive_start_radius;
	min_mg_dist = adaptive_start_point_distance;

	T min_mg_dist_pow_2;
	size_t prev_node_combinations_radius = -1;

	while (1)	// loop to decrease min_radius
	{
		// initialize the minimum allowed distance
		min_mg_dist = adaptive_start_point_distance;

		size_t node_combinations = 0;	// found node combinations for graph1
		size_t prev_node_combinations = -1;	// found node combinations for graph1

		// iterate over all possible combinations combining 4 points of the first graph
		// compute min_mg_dist
		//
		// this method could be implemented faster using a histogram like integral over a discrete graph giving the matchings for the distance
		typename list<CMatchPoint>::iterator iter1, iter2, iter3, iter4;	// iterators for marker points
		while (min_mg_dist > min_matrix_generation_point_distance)
		{
			min_mg_dist_pow_2 = min_mg_dist * min_mg_dist;
			node_combinations = 0;

			/**
			 * compute number of nodes from graph 1 within the allowed parameters
			 * !!!
			 * we count the possible matchings as follows:
			 * only matchings are allowed with a distance between mutual points over min_mg_dist
			 */

			// we dont compute all combinations in real but only all combinatorial ways to choose 4 numbers n in [0;N] with n_1 < n_2 < n_3 < n_4
			iter1 = matchPoints.begin();
			for (; iter1 != matchPoints.end(); iter1++)
			{
				CPointRadius<3,T> &p1 = (*iter1).point;
				if (p1.radius < min_radius)	continue;

				iter2 = iter1;	iter2++;
				for (; iter2 != matchPoints.end(); iter2++)
				{
					CPointRadius<3,T> &p2 = (*iter2).point;
					if (p2.radius < min_radius)	continue;

					// check if distance to this points is leq than min_matrix_generation_point_distance
					// this avoids creating the matrix by points which are very near and thus produce a non accurate matrix
					if (p2.pos.dist2(p1.pos) < min_mg_dist_pow_2)
						continue;	// directly abort if the distance is below the allowed range

					iter3 = iter2;	iter3++;
					for (; iter3 != matchPoints.end(); iter3++)
					{
						CPointRadius<3,T> &p3 = (*iter3).point;
						if (p3.radius < min_radius)	continue;

						if (p3.pos.dist2(p1.pos) < min_mg_dist_pow_2)
							continue;	// directly abort if the distance is below the allowed range

						if (p3.pos.dist2(p2.pos) < min_mg_dist_pow_2)
							continue;	// directly abort if the distance is below the allowed range

						iter4 = iter3;	iter4++;
						for (; iter4 != matchPoints.end(); iter4++)
						{
							CPointRadius<3,T> &p4 = (*iter4).point;

							if (p4.radius < min_radius)	continue;

							if (p4.pos.dist2(p1.pos) < min_mg_dist_pow_2)
								continue;

							if (p4.pos.dist2(p2.pos) < min_mg_dist_pow_2)
								continue;

							if (p4.pos.dist2(p3.pos) < min_mg_dist_pow_2)
								continue;

							node_combinations++;
						}
					}
				}
			}

			// prev_node_combinations is used to output debug informations only if new combinations were found with decreasing min_dist
			if (adaptive_start_radius == 0)	// output only informations if the min_radius is not computed adaptive
			if (prev_node_combinations != node_combinations)
			{
				std::cout << "G: min_dist = " << min_mg_dist << ", found_combinations: " << node_combinations << std::endl;
				prev_node_combinations = node_combinations;
			}

			// decrease the minimum distance for the next round
			min_mg_dist--;

			// if enough combinations were found, break
			if (adaptive_max_points < node_combinations)
				break;
		}

		// no adaptive minimum radius
		if (adaptive_start_radius == 0)
			break;

		// output debug information for current radius
		if (prev_node_combinations_radius != node_combinations)
		{
			std::cout << "G: (r=" << min_radius << ") min_dist = " << min_mg_dist << ", found_combinations: " << node_combinations << std::endl;
			prev_node_combinations_radius = node_combinations;
		}

		// check again if enough combinations were found and break.
		// otherwise continue the procedure to find the minimum radius
		// because not enough points were found within the minimum distance min_matrix_generation_point_distance
		if (adaptive_max_points < node_combinations)
		{
			min_radius++;
			min_mg_dist++;
			break;
		}

		// min_radius got too small => break
		if (min_radius <= 0)
			break;

		min_radius--;
	}
}



/**
 * this function can be called by a thread
 * \param thread_id	id of thread for load distribution
 */
void search_best_matching(
	list< CMatchPoint > &matchPoints1,
	T min_radius1,
	T min_mg1_dist_pow_2,
	T inv_match_points1_count,

	list< CMatchPoint > &matchPoints2,
	T min_radius2,
	T min_mg2_dist_pow_2,
	T inv_match_points2_count,

	T *min_distance,
	mat4f *best_matrix_trans,
	size_t *found_valid_matrices_count,

	int thread_id = 0
)
{
	CVector<3,T> rotation;
	CVector<3,T> scaling;
	CVector<3,T> shearing;
	CVector<3,T> translation;

	CMatrix4<T> matrix_trans;	// temporary matrix to store the computed transformation matrix

	int enumeration = 0;		// enumeration helper for load distribution
	int enumeration_dots = 0;	// enumerator for dot status output

	// iterate over all possible combinations combining 4 points of the first graph
	typename list<CMatchPoint>::iterator iter11 = matchPoints1.begin();
	for (; iter11 != matchPoints1.end(); iter11++)
	{
		CPointRadius<3,T> &p11 = (*iter11).point;
		if (p11.radius < min_radius1)	continue;

		typename list<CMatchPoint>::iterator iter12 = iter11;	// copy iterator for next for loop
		iter12++;
		if (iter12 == matchPoints1.end())	break;		// break if we reached the end (no more matchings)

		for (; iter12 != matchPoints1.end(); iter12++)
		{
			CPointRadius<3,T> &p12 = (*iter12).point;
			if (p12.radius < min_radius1)	continue;

			// check if distance to this points is leq than min_matrix_generation_point_distance
			// this avoids creating the matrix by points which are very near and thus produce a non accurate matrix
			if (p12.pos.dist2(p11.pos) < min_mg1_dist_pow_2)
				continue;

			typename list<CMatchPoint>::iterator iter13 = iter12;
			iter13++;
			if (iter13 == matchPoints1.end())	break;

			for (; iter13 != matchPoints1.end(); iter13++)
			{
				CPointRadius<3,T> &p13 = (*iter13).point;

				if (p13.radius < min_radius1)	continue;

				if (p13.pos.dist2(p11.pos) < min_mg1_dist_pow_2)
					continue;
				if (p13.pos.dist2(p12.pos) < min_mg1_dist_pow_2)
					continue;

				typename list<CMatchPoint>::iterator iter14 = iter13;
				iter14++;
				if (iter14 == matchPoints1.end())	break;

				for (; iter14 != matchPoints1.end(); iter14++)
				{
					CPointRadius<3,T> &p14 = (*iter14).point;

					if (p14.radius < min_radius1)	continue;

					if (p14.pos.dist2(p11.pos) < min_mg1_dist_pow_2)
						continue;
					if (p14.pos.dist2(p12.pos) < min_mg1_dist_pow_2)
						continue;
					if (p14.pos.dist2(p13.pos) < min_mg1_dist_pow_2)
						continue;

					// use threads
					if (number_of_threads > 1)
					{
						enumeration++;

						if (enumeration >= number_of_threads)
							enumeration = 0;

						if (enumeration != thread_id)
							continue;		// continue, if this matching is handled by another thread
					}

					/**
					 * MATCHING POINTS FOR EDGE1 FOUND
					 * we continue by finding all possible combinations on edge2!
					 */

					typename list<CMatchPoint>::iterator iter21 = matchPoints2.begin();
					for (; iter21 != matchPoints2.end(); iter21++)
					{
						CPointRadius<3,T> &p21 = (*iter21).point;
						if (p21.radius < min_radius2)	continue;

						// if the 2 matching points have a too large distance (untransformed), it probably is not the right matching
						if (p21.pos.dist2(p11.pos) > max_untransformed_distance_pow_2)
							continue;

						// for combinatorial reasons (to match also p_1 to p_4), we have to start at begin
						typename list<CMatchPoint>::iterator iter22 = matchPoints2.begin();

						for (; iter22 != matchPoints2.end(); iter22++)
						{
							CPointRadius<3,T> &p22 = (*iter22).point;

							if (p22.radius < min_radius2)	continue;

							// choosing the same points (iter21 == iter22) would abort here if min_matrix_generation_point_distance_pow_2 > 0
							if (p22.pos.dist2(p21.pos) < min_mg2_dist_pow_2)
								continue;

							if (p22.pos.dist2(p12.pos) > max_untransformed_distance_pow_2)
								continue;

							typename list<CMatchPoint>::iterator iter23 = matchPoints2.begin();

							for (; iter23 != matchPoints2.end(); iter23++)
							{
								CPointRadius<3,T> &p23 = (*iter23).point;

								if (p23.radius < min_radius2)	continue;

								if (p23.pos.dist2(p21.pos) < min_mg2_dist_pow_2)
									continue;

								if (p23.pos.dist2(p22.pos) < min_mg2_dist_pow_2)
									continue;

								if (p23.pos.dist2(p13.pos) > max_untransformed_distance_pow_2)
									continue;

								typename list<CMatchPoint>::iterator iter24 = matchPoints2.begin();

								for (; iter24 != matchPoints2.end(); iter24++)
								{
									CPointRadius<3,T> &p24 = (*iter24).point;

									if (p24.radius < min_radius2)	continue;

									if (p24.pos.dist2(p21.pos) < min_mg2_dist_pow_2)
										continue;
									if (p24.pos.dist2(p22.pos) < min_mg2_dist_pow_2)
										continue;
									if (p24.pos.dist2(p23.pos) < min_mg2_dist_pow_2)
										continue;

									if (p24.pos.dist2(p14.pos) > max_untransformed_distance_pow_2)
										continue;

//									std::cout << p21.pos << p22.pos << p23.pos << p24.pos;
									// TODO: improve by reusing already solved equation systems
									if (!compute_matrix(	matrix_trans,
												*iter11, *iter12, *iter13, *iter14,
												*iter21, *iter22, *iter23, *iter24	)
									)
										continue;

									if (!decompose_and_filter_matrix(matrix_trans, rotation, scaling, shearing, translation))
										continue;

									(*found_valid_matrices_count)++;

									// ok, now we have a valid transformation matrix with values within the given borders
									// now we check the distance from the points in graph1 the other points from graph2 transformed by the matrix
									/*
									dfile << matrix_trans;

									for (	typename list<CMatchPoint>::iterator iterx = matchPoints1.begin();
										iterx != matchPoints1.end();
									       	iterx++)
									{
										CPointRadius<3,T> &px = (*iterx).point;
										dfile << px.pos << px.radius;
									}
									dfile << std::endl;


									for (	typename list<CMatchPoint>::iterator iterx = matchPoints2.begin();
										iterx != matchPoints2.end();
									       	iterx++)
									{
										CPointRadius<3,T> &px = (*iterx).point;
										dfile << px.pos << px.radius;
									}
									dfile << std::endl;
									dfile << min_distance << std::endl;
									*/

									T new_min_distance = compute_minimum_distance(
													matrix_trans,
													matchPoints1,
													tree2,
													*min_distance,
													inv_match_points1_count
												);
//									dfile << "x: " << new_min_distance << " " << min_distance << endl;

									if (use_mutual_graph_distance)
									{
										// also match graph2 to graph1!
										CMatrix4<T> inv_matrix_trans;
										if (matrix_trans.compute_inverse(inv_matrix_trans))
										{
											// compute inverse matrix
											new_min_distance += compute_minimum_distance(
															inv_matrix_trans,
														       	matchPoints2,
														       	tree1,
														       	*min_distance,
															inv_match_points2_count
														);

											if (*min_distance > new_min_distance)
											{
												*min_distance = new_min_distance;
												*best_matrix_trans = matrix_trans;
											}
										}
									}
									else
									{
										if (*min_distance > new_min_distance)
										{
//											dfile << found_valid_matrices_count << ": " << new_min_distance << endl;
											*min_distance = new_min_distance;
											*best_matrix_trans = matrix_trans;
										}
									}
								}
							}
						}
					}
				}
			}

			if (enumeration_dots == thread_id || number_of_threads <= 1)
				std::cout << "." << std::flush;
		}
		if (number_of_threads > 1)
		{
			enumeration_dots++;

			if (enumeration_dots >= number_of_threads)
				enumeration_dots = 0;

			if (enumeration_dots == thread_id)
				std::cout << "|" << std::flush;
		}
		else
		{
			std::cout << "|" << std::flush;
		}
	}
}


struct s_thread_parameters
{
	int id;

	list< CMatchPoint > *matchPoints1;
	T min_radius1;
	T min_mg1_dist_pow_2;
	T inv_match_points1_count;

	list< CMatchPoint > *matchPoints2;
	T min_radius2;
	T min_mg2_dist_pow_2;
	T inv_match_points2_count;

	T min_distance;
	mat4f best_matrix_trans;
	size_t found_valid_matrices_count;

	CFilterMatchingSimple<T> *filterClass;
};

/**
 * startpoint for pthread
 */
static void *search_best_matching_thread(void *a)
{
	struct s_thread_parameters *p = (struct s_thread_parameters *)a;

	p->filterClass->search_best_matching(	*(p->matchPoints1), p->min_radius1, p->min_mg1_dist_pow_2, p->inv_match_points1_count,
						*(p->matchPoints2), p->min_radius2, p->min_mg2_dist_pow_2, p->inv_match_points2_count,
						&(p->min_distance),
						&(p->best_matrix_trans),
						&(p->found_valid_matrices_count),
						p->id
			);

	return NULL;
}

public:
/**
 * call the filter and store best matching matrix to best_matrix_trans
 *
 * \param p_graph1	reference graph
 * \param p_graph2	destination graph
 * \param number_of_threads	threads to use for matching. set to 1 to disable threading
 */
void filter(CGraph<3,T> &p_graph1, CGraph<3,T> &p_graph2)
{
	CGraph<3,T> *graph1;
	CGraph<3,T> *graph2;

	// graph1 should be the one with less or equal edges than graph2 to optimize matching
	graph1 = &p_graph1;
	graph2 = &p_graph2;

	std::cout << "." << std::flush;
	std::cout << "initializing matching test points..." << std::endl;

	// initialize edge values for graph1
	list< CMatchPoint > matchPoints1;
	graph1->recompEdgeLengths();
	for (typename list<typename CGraph<3,T>::CEdge>::iterator i1 = graph1->edges.begin(); i1 != graph1->edges.end(); i1++)
	{
		typename CGraph<3,T>::CEdge &edge1 = (*i1);

		for (typename list<typename CGraph<3,T>::CNode>::iterator n1 = edge1.nodes.begin(); n1 != edge1.nodes.end(); n1++)
		{
			matchPoints1.push_back(CMatchPoint((*n1).pos - pre_translation, (*n1).radius, edge1, n1));
			std::cout << matchPoints1.back().point.pos << " (" << matchPoints1.back().point.radius << ")\t";
		}
	}

	// sort matching points by radius to help the early termination while computing the distance of both graphs for the current matrix
	// usind large radii first should create earlier termination because using more characteristic marker points
	matchPoints1.sort(CMatchPoint::cmp_sorted_points);

	std::cout << std::endl;
	std::cout << "match points for graph1: " << matchPoints1.size() << std::endl;

	// initialize edge values for graph2
	list< CMatchPoint > matchPoints2;
	graph2->recompEdgeLengths();
	for (typename list<typename CGraph<3,T>::CEdge>::iterator i2 = graph2->edges.begin(); i2 != graph2->edges.end(); i2++)
	{
		typename CGraph<3,T>::CEdge &edge2 = (*i2);

		for (typename list<typename CGraph<3,T>::CNode>::iterator n2 = edge2.nodes.begin(); n2 != edge2.nodes.end(); n2++)
		{
			matchPoints2.push_back(CMatchPoint((*n2).pos - pre_translation, (*n2).radius, edge2, n2));
			std::cout << matchPoints2.back().point.pos << " (" << matchPoints2.back().point.radius << ")\t";
		}
	}

	std::cout << std::endl;
	std::cout << "match points for graph2: " << matchPoints2.size() << std::endl;

	/**********************************************************
	 * insert point list into tree and to sorted point list
	 */
	std::cout << " kdtree1(" << flush;
	std::cout << "." << std::flush;
	for (typename list< CMatchPoint >::iterator iter = matchPoints1.begin(); iter != matchPoints1.end(); iter++)
		tree1->insert(&(*iter));

	std::cout << "." << std::flush;
	tree1->optimize();		// optimize the tree1
	std::cout << tree1->size() << std::endl;
	std::cout << ")" << std::endl << std::flush;

	std::cout << " kdtree2(" << flush;
	std::cout << "." << std::flush;
	for (typename list< CMatchPoint >::iterator iter = matchPoints2.begin(); iter != matchPoints2.end(); iter++)
		tree2->insert(&(*iter));

	std::cout << "." << std::flush;
	tree2->optimize();		// optimize the tree2
	std::cout << tree2->size() << ")" << std::endl << std::flush;

	std::cout << "matching..." << std::endl;

	T min_mg1_dist_pow_2;
	T min_mg2_dist_pow_2;

	T min_mg1_dist;
	T min_mg2_dist;

	T min_radius1;
	T min_radius2;

	if (adaptive_max_points > 0)
	{
		/*
		 * information for adaptive control:
		 *
		 * we search for a value min_mg?_dist for each graph reducing the matching possibilities
		 * the value is given by the maximum number of allowed points to match for the distance,
		 * therefore we check all possible combinations until such a border value was found
		 *
		 * TODO: use adaption on the fly (and a better method)
		 */
		filter_helper_get_adaptive_values(matchPoints1, min_mg1_dist, min_radius1);
		std::cout << "G1: min_dist = " << min_mg1_dist << ", min_radius = " << min_radius1 << std::endl;;

		filter_helper_get_adaptive_values(matchPoints2, min_mg2_dist, min_radius2);
		std::cout << "G2: min_dist = " << min_mg2_dist << ", min_radius = " << min_radius2 << std::endl;;
	}
	else
	{
		// no adaptive control -> use given value
		min_mg1_dist = min_matrix_generation_point_distance;
		min_mg2_dist = min_matrix_generation_point_distance;

		min_radius1 = 0;
		min_radius2 = 0;
	}

	min_mg1_dist_pow_2 = min_mg1_dist * min_mg1_dist;
	min_mg2_dist_pow_2 = min_mg2_dist * min_mg2_dist;

	std::cout << "G1,G2: final min_dist = (" << min_mg1_dist << ", " << min_mg2_dist << ")" << std::endl;


	// parameters of basic matrices
	CVector<3,T> rotation;
	CVector<3,T> scaling;
	CVector<3,T> shearing;
	CVector<3,T> translation;

	CMatrix4<T> matrix_trans;	// temporary matrix to store the computed transformation matrix
//
	T match_points1_count = matchPoints1.size();
	T inv_match_points1_count = (T)1/match_points1_count;

	T match_points2_count = matchPoints2.size();
	T inv_match_points2_count = (T)1/match_points2_count;

	size_t found_valid_matrices_count = 0;

	if (number_of_threads <= 1)
	{
		std::cout << "(no threading)" << std::endl;

		min_distance	= std::numeric_limits<T>::max();

		// no threading
		std::cout << "[" << std::flush;
		search_best_matching(	matchPoints1, min_radius1, min_mg1_dist_pow_2, inv_match_points1_count,
					matchPoints2, min_radius2, min_mg2_dist_pow_2, inv_match_points2_count,
					&min_distance,
					&best_matrix_trans,
					&found_valid_matrices_count,
					0
				);
		std::cout << "]" << std::endl;
	}
	else
	{
		std::cout << "(using " << number_of_threads << " threads)" << std::endl;
		std::cout << "[" << std::flush;

		struct s_thread_parameters thread_parameters[number_of_threads];
		pthread_t threads[number_of_threads];

		for (int i = 0; i < number_of_threads; i++)
		{
			thread_parameters[i].id = i;

			thread_parameters[i].matchPoints1	= &matchPoints1;
			thread_parameters[i].min_radius1	= min_radius1;
			thread_parameters[i].min_mg1_dist_pow_2	= min_mg1_dist_pow_2;
			thread_parameters[i].inv_match_points1_count	= inv_match_points1_count;

			thread_parameters[i].matchPoints2	= &matchPoints2;
			thread_parameters[i].min_radius2	= min_radius2;
			thread_parameters[i].min_mg2_dist_pow_2	= min_mg2_dist_pow_2;
			thread_parameters[i].inv_match_points2_count	= inv_match_points2_count;

			thread_parameters[i].min_distance	= std::numeric_limits<T>::max();
			thread_parameters[i].best_matrix_trans.loadIdentity();
			thread_parameters[i].found_valid_matrices_count	= 0;

			thread_parameters[i].filterClass = this;

			if (pthread_create(&(threads[i]), NULL, &search_best_matching_thread, &(thread_parameters[i])) != 0)
				std::cerr << "pthread_create FAILED" << std::endl;
		}

		min_distance	= std::numeric_limits<T>::max();
		for (int i = 0; i < number_of_threads; i++)
		{
			if (pthread_join(threads[i], 0) != 0)
			{
				std::cerr << "pthread_join FAILED" << std::endl;
				continue;
			}

			found_valid_matrices_count += thread_parameters[i].found_valid_matrices_count;

			if (min_distance > thread_parameters[i].min_distance)
			{
				min_distance = thread_parameters[i].min_distance;
				best_matrix_trans = thread_parameters[i].best_matrix_trans;
			}
		}
		std::cout << "]" << std::endl;
	}

	std::cout << "found valid matching matrices: " << found_valid_matrices_count << std::endl;

	if (min_distance == std::numeric_limits<T>::max())
	{
		std::cerr << "no matching found" << std::endl;
		clear_tree1();
		clear_tree2();
	
		best_pre_translated_matrix_trans.loadIdentity();
		return;
	}

	if (use_mutual_graph_distance)
		min_distance *= (T)0.5;

	std::cout << " min_distance: " << min_distance << std::endl;
	std::cout << std::endl;

	// muliply pre translation on matrix
	CMatrix4<T> transl1;
	transl1.loadIdentity();
	transl1[0][3] = -pre_translation[0];
	transl1[1][3] = -pre_translation[1];
	transl1[2][3] = -pre_translation[2];

	CMatrix4<T> transl2;
	transl2.loadIdentity();
	transl2[0][3] = +pre_translation[0];
	transl2[1][3] = +pre_translation[1];
	transl2[2][3] = +pre_translation[2];

	best_pre_translated_matrix_trans = transl2 * best_matrix_trans * transl1;

	clear_tree1();
	clear_tree2();
}

};

#endif
