/*
 * This file is part of the segmentation and registration toolkit (SphereSegRegTk).
 *
 * SphereSegRegTk is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * SphereSegRegTk is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with SphereSegRegTk.  If not, see <http://www.gnu.org/licenses/>.
 */
#ifndef __CVECTOR_HH
	#error "dont include CVector3.hh directly!"
#endif

#ifndef __CVECTOR2_HH
#define __CVECTOR2_HH
/*
 * CHANGELOG:
 *
 * 2008-02-27: martin schreiber
 * 	dist2();
 *
 * 2008-10-22: martin schreiber
 * 	made some modifications to use generic vector CVector
 *
 */

#include <iostream>
#include "CMath.hh"

/// 2D Vector handler
template <typename T>
class CVector<2,T>
{
public:
	T data[2];	///< vector data

	inline CVector()
	{
		setZero();
	}

	/**
	 * initialize vector with (x0, x1)
	 */
	inline CVector(const T x0, const T x1)
	{
		data[0] = x0;
		data[1] = x1;
	}

	/**
	 * initialize all vector components with the scalar value 'x'
	 */
	inline CVector(const T x)
	{
		data[0] = x;
		data[1] = x;
	}

	/**
	 * initialize vector components with the array 'v'
	 */
	inline CVector(const T v[2])
	{
		data[0] = v[0];
		data[1] = v[1];
	}

	/**
	 * set all components of vector to 0
	 */
	inline void setZero()
	{
		data[0] = T(0);
		data[1] = T(0);
	}

	/**
	 * return normalized (length=1) vector
	 */
	inline CVector getNormal()
	{
		CVector<2,T> v = *this;
		float inv_length = 1.0f/sqrtf(v[0]*v[0] + v[1]*v[1]);
		v[0] *= inv_length;
		v[1] *= inv_length;
		return v;
	}

	/**
	 * compute and return the dot product of this vector and 'v'
	 * \return dot product
	 */
	inline T getDotProd(const CVector<2,T> &v)
	{
		return v.data[0]*data[0] + v.data[1]*data[1];
	}

	/**
	 * compute and return the cross product of this vector and 'a'
	 * \return cross product
	 */
	inline T getCrossProd(CVector<2,T> &a)
	{
		return data[0]*a.data[1] - data[1]*a.data[0];
	}


	/**
	 * return elements in cube.
	 * e.g. if the vector components give the resolution of a cube, this function returns the number of cells
	 */
	inline T elements()
	{
		return (data[0]*data[1]);
	}

	/**
	 * return length of vector
	 */
	inline T length()
	{
		return CMath<T>::sqrt(data[0]*data[0] + data[1]*data[1]);
	}

	/**
	 * return (length*length) of vector
	 */
	inline T length2()
	{
		return data[0]*data[0] + data[1]*data[1];
	}

	/**
	 * return square distance to other point
	 */
	inline T dist2(const CVector<2,T> &v)
	{
		CVector<2,T> d = CVector<2,T>(v.data[0] - data[0], v.data[1] - data[1]);
		return (d[0]*d[0] + d[1]*d[1]);
	}

	/**
	 * return distance to other point
	 */
	inline T dist(const CVector<2,T> &v)
	{
		CVector<2,T> d = CVector<2,T>(v.data[0] - data[0], v.data[1] - data[1]);
		return sqrtf(d[0]*d[0] + d[1]*d[1]);
	}

	/**
	 * normalize the vector
	 */
	inline void normalize()	
	{
		T il = 1/length();
		data[0] *= il;
		data[1] *= il;
	}

	/**
	 * clamp to values -1 and 1
	 */
	inline void clamp1_1()
	{
		data[0] = (data[0] < -1 ? -1 : (data[0] > 1 ? 1 : data[0]));
		data[1] = (data[1] < -1 ? -1 : (data[1] > 1 ? 1 : data[1]));
	}

	/*******************
	 * OPERATORS
	 *******************/
	inline CVector<2,T>&	operator=(const T a[2])	{	data[0] = a[0]; data[1] = a[1]; return *this;	};
	inline CVector<2,T>&	operator=(CVector<2,T> const& a)	{	data[0] = a.data[0]; data[1] = a.data[1]; return *this;	};

	inline CVector<2,T>	operator-()	{	return CVector<2,T>(-data[0], -data[1]);	}

// T
	inline CVector<2,T>	operator+(const T a)	{	return CVector<2,T>(data[0]+a, data[1]+a);	}
	inline CVector<2,T>	operator-(const T a)	{	return CVector<2,T>(data[0]-a, data[1]-a);	}
	inline CVector<2,T>	operator*(const T a)	{	return CVector<2,T>(data[0]*a, data[1]*a);	}
	inline CVector<2,T>	operator/(const T a)	{	return CVector<2,T>(data[0]/a, data[1]/a);	}
	inline CVector<2,T>&	operator+=(const T a)	{	data[0] += a; data[1] += a;	return *this;	}
	inline CVector<2,T>&	operator-=(const T a)	{	data[0] -= a; data[1] -= a;	return *this;	}
	inline CVector<2,T>&	operator*=(const T a)	{	data[0] *= a; data[1] *= a;	return *this;	}
	inline CVector<2,T>&	operator/=(const T a)	{	data[0] /= a; data[1] /= a;	return *this;	}

// CVector
	inline CVector<2,T>	operator+(const CVector<2,T> &v)	{	return CVector<2,T>(data[0]+v.data[0], data[1]+v.data[1]);	}
	inline CVector<2,T>	operator-(const CVector<2,T> &v)	{	return CVector<2,T>(data[0]-v.data[0], data[1]-v.data[1]);	}
	inline CVector<2,T>	operator*(const CVector<2,T> &v)	{	return CVector<2,T>(data[0]*v.data[0], data[1]*v.data[1]);	}
	inline CVector<2,T>	operator/(const CVector<2,T> &v)	{	return CVector<2,T>(data[0]/v.data[0], data[1]/v.data[1]);	}

	inline CVector<2,T>&	operator+=(const CVector<2,T> &v)	{	data[0] += v.data[0]; data[1] += v.data[1]; 	return *this;	}
	inline CVector<2,T>&	operator-=(const CVector<2,T> &v)	{	data[0] -= v.data[0]; data[1] -= v.data[1]; 	return *this;	}

	inline bool	operator==(const CVector<2,T> &v)	{	return bool(data[0] == v.data[0] && data[1] == v.data[1]);	}
	inline bool	operator!=(const CVector<2,T> &v)	{	return bool(data[0] != v.data[0] || data[1] != v.data[1]);	}


	inline T& operator[](const int i)
	{
		return data[i];
	}

	/// compare set for sort operation
	struct compareSet
	{

		inline bool operator()(CVector<2,float> *v1, CVector<2,float> *v2)
		{
			if ((*v1)[0] != (*v2)[0])
				return (*v1)[0] < (*v2)[0];
			if ((*v1)[1] != (*v2)[1])
				return (*v1)[1] < (*v2)[1];
			return false;
		}

		inline bool operator()(const CVector<2,T> &v1, const CVector<2,T> &v2)
		{
			if (v1.data[0] != v2.data[0])
				return v1.data[0] < v2.data[0];
			if (v1.data[1] != v2.data[1])
				return v1.data[1] < v2.data[1];
			return false;
		}
	};
};


typedef CVector<2,float>	vec2f;
typedef CVector<2,int>		vec2i;


template <class T>
inline
std::ostream&
operator<<(std::ostream &co, CVector<2,T> &v)
{
	return co << "[" << v[0] << ", " << v[1] << "]";
}


#endif
