/*
 * This file is part of the segmentation and registration toolkit (SphereSegRegTk).
 *
 * SphereSegRegTk is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * SphereSegRegTk is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with SphereSegRegTk.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef __CVECTOR_HH
	#error "dont include CVector4.hh directly!"
#endif

#ifndef __CVECTOR4_HH
#define __CVECTOR4_HH
/**
 * CHANGELOG:
 *
 * 2008-02-27: martin schreiber
 * 	dist2();
 *
 */

#include <iostream>
#include "CMath.hh"

/// 4D Vector handler
template <typename T>
class CVector<4,T>
{
public:
	// vector data
	T data[4];

	/*******************
	 * CONSTRUCTURS
	 *******************/
	inline CVector()
	{
		setZero();
	}

	/**
	 * initialize vector with (x0, x1, x2, x3)
	 */
	inline CVector(const T x0, const T x1, const T x2, const T x3)
	{
		data[0] = x0;
		data[1] = x1;
		data[2] = x2;
		data[3] = x3;
	}

	/**
	 * initialize all vector components with the scalar value 'x'
	 */
	inline CVector(const T x)
	{
		data[0] = x;
		data[1] = x;
		data[2] = x;
		data[3] = x;
	}

	/**
	 * initialize integer vector components with the array 'v' converting to the type (T)
	 */
	inline CVector(const T v[4])
	{
		data[0] = v[0];
		data[1] = v[1];
		data[2] = v[2];
		data[3] = v[3];
	}


	/**
	 * set all components of vector to 0
	 */
	inline void setZero()
	{
		data[0] = T(0);
		data[1] = T(0);
		data[2] = T(0);	
		data[3] = T(0);	
	}

	/**
	 * return length of vector
	 */
	inline T length()
	{
		return CMath<T>::sqrt(data[0]*data[0] + data[1]*data[1] + data[2]*data[2] + data[3]*data[3]);
	}

	/**
	 * return (length*length) of vector
	 */
	inline T length2()
	{
		return data[0]*data[0] + data[1]*data[1] + data[2]*data[2] + data[3]*data[3];
	}



	/*******************
	 * OPERATORS
	 *******************/
	inline CVector<4,T>&	operator=(const T a[4])	{	data[0] = a[0]; data[1] = a[1]; data[2] = a[2];	data[3] = a[3]; return *this;	};
	inline CVector<4,T>&	operator=(CVector<4,T> const & a)	{	data[0] = a.data[0]; data[1] = a.data[1]; data[2] = a.data[2];	data[3] = a.data[3]; return *this;	};
	inline CVector<4,T>&	operator=(CVector<3,T> const & a)	{	data[0] = a.data[0]; data[1] = a.data[1]; data[2] = a.data[2];	data[3] = 1; return *this;	};


// T
	inline CVector<4,T>	operator+(const T a)	{	return CVector<4,T>(data[0]+a, data[1]+a, data[2]+a, data[3]+a);	}
	inline CVector<4,T>	operator-(const T a)	{	return CVector<4,T>(data[0]-a, data[1]-a, data[2]-a, data[3]-a);	}
	inline CVector<4,T>	operator*(const T a)	{	return CVector<4,T>(data[0]*a, data[1]*a, data[2]*a, data[3]*a);	}
	inline CVector<4,T>	operator/(const T a)	{	return CVector<4,T>(data[0]/a, data[1]/a, data[2]/a, data[3]/a);	}
	inline CVector<4,T>&	operator+=(const T a)	{	data[0] += a; data[1] += a; data[2] += a; data[3] += a;	return *this;	}
	inline CVector<4,T>&	operator-=(const T a)	{	data[0] -= a; data[1] -= a; data[2] -= a; data[3] -= a;	return *this;	}
	inline CVector<4,T>&	operator*=(const T a)	{	data[0] *= a; data[1] *= a; data[2] *= a; data[3] *= a;	return *this;	}
	inline CVector<4,T>&	operator/=(const T a)	{	data[0] /= a; data[1] /= a; data[2] /= a; data[3] /= a;	return *this;	}

	inline CVector<4,T>	operator+(const CVector<4,T> &v)	{	return CVector<4,T>(data[0]+v.data[0], data[1]+v.data[1], data[2]+v.data[2], data[3]+v.data[3]);	}
	inline CVector<4,T>	operator-(const CVector<4,T> &v)	{	return CVector<4,T>(data[0]-v.data[0], data[1]-v.data[1], data[2]-v.data[2], data[3]-v.data[3]);	}
	inline CVector<4,T>	operator*(const CVector<4,T> &v)	{	return CVector<4,T>(data[0]*v.data[0], data[1]*v.data[1], data[2]*v.data[2], data[3]*v.data[3]);	}
	inline CVector<4,T>	operator/(const CVector<4,T> &v)	{	return CVector<4,T>(data[0]/v.data[0], data[1]/v.data[1], data[2]/v.data[2], data[3]/v.data[3]);	}

	inline CVector<4,T>&	operator+=(const CVector<4,T> &v)	{	data[0] += v.data[0]; data[1] += v.data[1]; data[2] += v.data[2]; data[3] += v.data[3];	return *this;	}
	inline CVector<4,T>&	operator-=(const CVector<4,T> &v)	{	data[0] -= v.data[0]; data[1] -= v.data[1]; data[2] -= v.data[2]; data[3] -= v.data[3];	return *this;	}

	inline bool	operator==(const CVector<4,T> &v)	{	return bool(data[0] == v.data[0] && data[1] == v.data[1] && data[2] == v.data[2] && data[3] == v.data[3]);	}
	inline bool	operator!=(const CVector<4,T> &v)	{	return bool(data[0] != v.data[0] || data[1] != v.data[1] || data[2] != v.data[2] || data[3] != v.data[3]);	}


	inline T& operator[](const int i)
	{
		return data[i];
	}

	/// compare set for sort operation
	struct compareSet
	{

		inline bool operator()(CVector<4,float> *v1, CVector<4,float> *v2)
		{
			if ((*v1)[0] != (*v2)[0])
				return (*v1)[0] < (*v2)[0];
			if ((*v1)[1] != (*v2)[1])
				return (*v1)[1] < (*v2)[1];
			if ((*v1)[2] != (*v2)[2])
				return (*v1)[2] < (*v2)[2];
			if ((*v1)[3] != (*v2)[3])
				return (*v1)[3] < (*v2)[3];
			return false;
		}

		inline bool operator()(const CVector<4,T> &v1, const CVector<4,T> &v2)
		{
			if (v1.data[0] != v2.data[0])
				return v1.data[0] < v2.data[0];
			if (v1.data[1] != v2.data[1])
				return v1.data[1] < v2.data[1];
			if (v1.data[2] != v2.data[2])
				return v1.data[2] < v2.data[2];
			if (v1.data[3] != v2.data[3])
				return v1.data[3] < v2.data[3];
			return false;
		}
	};
};


typedef CVector<4,float>	vec4f;
typedef CVector<4,int>	vec4i;


template <class T>
inline
std::ostream&
operator<<(std::ostream &co, CVector<4,T> &v)
{
	return co << "[" << v[0] << ", " << v[1] << ", " << v[2] << ", " << v[3] << "]";
}


#endif
